"""
Module for the Coherent Spatio-Temporal Scale Separation with DMD.

References:
- Dylewsky, D., Tao, M., & Kutz, J. N. (2019). Dynamic mode decomposition for
multiscale nonlinear physics. Physics Review E, 99(6),
10.1103/PhysRevE.99.063311. https://doi.org/10.1103/PhysRevE.99.063311
"""

import copy

import numpy as np
from sklearn.cluster import MiniBatchKMeans
from sklearn.metrics import silhouette_score
import matplotlib.pyplot as plt
import xarray as xr

from pydmd.bopdmd import BOPDMD
from .utils import compute_rank, compute_svd


class COSTS:
    """Coherent Spatio-Temporal Scale Separation with DMD.

    :param n_components: Number of independent frequency bands for this
        window length.
    :type n_components: int
    :param svd_rank: The rank of the BOPDMD fit.
    :type svd_rank: int
    :param global_svd: Flag indicating whether to find the proj_basis and
        initial values using the entire dataset instead of individually for
        each window. Generally using the global_svd speeds up the fitting
        process by not finding a new initial value for each window. Default
        is True.
    :type global_svd: bool
    :param initialize_artificially: Flag indicating whether to initialize the
        DMD using imaginary eigenvalues (i.e., the imaginary component of the
        cluster results from a previous iteration) through the
        `cluster_centroids` keyword. Default is False.
    :type initialize_artificially: bool
    :param pydmd_kwargs: Keyword arguments to pass onto the BOPDMD object.
    :type pydmd_kwargs: dict
    :param cluster_centroids: Cluster centroids from a previous fitting
        iteration to use for the initial guess of the eigenvalues. Should
        only be the imaginary component.
    :type cluster_centroids: numpy array
    :param reset_alpha_init: Flag indicating whether the initial guess for
        the BOPDMD eigenvalues should be reset for each window. Resetting the
        initial value increases the computation time due to finding a new
        initial guess. Default is False.
    :type reset_alpha_init: bool
    :param force_even_eigs: Flag indicating whether an even svd_rank should be
        forced when not specifying the svd_rank directly (i.e., svd_rank=0).
        Default is True.
    :type global_svd: bool
    :param max_rank: Maximum svd_rank allowed when the svd_rank is found
        through rank truncation (i.e., svd_rank=0).
    :type max_rank: int
    :param init_alpha: Initial guess for the eigenvalues provided to BOPDMD.
        Must be equal to the `svd_rank`.
    :type init_alpha: numpy array
    :param max_rank: Maximum allowed `svd_rank`. Overrides the optimal rank
        truncation if `svd_rank=0`.
    :type max_rank: int
    :param n_components: Number of frequency bands to use for clustering.
    :type n_components: int
    :param force_even_eigs: Flag specifying if the `svd_rank` should be forced
        to be even.
    :type force_even_eigs: bool
    :param reset_alpha_init: Flag specifying if the initial eigenvalue guess
        should be reset between windows.
    :type reset_alpha_init: bool
    """

    def __init__(
        self,
        svd_rank=None,
        global_svd=True,
        initialize_artificially=False,
        use_last_freq=False,
        init_alpha=None,
        pydmd_kwargs=None,
        cluster_centroids=None,
        reset_alpha_init=False,
        force_even_eigs=True,
        max_rank=None,
        n_components=None,
    ):
        self._hist_kwargs = None
        self._omega_label = None
        self._step_size = None
        self._window_length = None
        self._n_components = n_components
        self._svd_rank = svd_rank
        self._global_svd = global_svd
        self._initialize_artificially = initialize_artificially
        self._use_last_freq = use_last_freq
        self._init_alpha = init_alpha
        self._cluster_centroids = cluster_centroids
        self._force_even_eigs = force_even_eigs
        self._max_rank = max_rank
        self._reset_alpha_init = reset_alpha_init

        # Initialize variables that are defined in fitting.
        self._n_data_vars = None
        self._n_time_steps = None
        self._window_length = None
        self._n_slides = None
        self._time_array = None
        self._modes_array = None
        self._omega_array = None
        self._amplitudes_array = None
        self._cluster_centroids = None
        self._omega_classes = None
        self._transform_method = None
        self._window_means_array = None
        self._non_integer_n_slide = None
        self._svd_rank_pre_allocate = None

        # Specify default keywords to hand to BOPDMD.
        if pydmd_kwargs is None:
            self._pydmd_kwargs = {
                "eig_sort": "imag",
                "proj_basis": None,
                "use_proj": False,
            }
        else:
            self._pydmd_kwargs = pydmd_kwargs
            self._pydmd_kwargs["eig_sort"] = pydmd_kwargs.get(
                "eig_sort", "imag"
            )
            self._pydmd_kwargs["proj_basis"] = pydmd_kwargs.get(
                "proj_basis", None
            )
            self._pydmd_kwargs["use_proj"] = pydmd_kwargs.get("use_proj", False)

    @property
    def svd_rank(self):
        """Return the svd_rank used for the BOPDMD fit.

        :return: the rank used for the svd truncation.
        :rtype: int or float
        """
        return self._svd_rank

    @property
    def global_svd(self):
        """Return if a global svd projection basis was used.

        :return: If a global svd was used for the BOP-DMD fit.
        :rtype: int or float
        """
        return self._global_svd

    @property
    def window_length(self):
        """Return the window length used.

        :return: the length of the windows used for this decomposition level.
        :rtype: int or float
        """
        return self._window_length

    @property
    def step_size(self):
        """Return the step size between each window.

        :return: the length of the windows used for this decomposition level.
        :rtype: int or float
        """
        return self._step_size

    @property
    def n_slides(self):
        """Return the number of slides performed for each window.

        :return: number of window slides for this decomposition level.
        :rtype: int
        """
        return self._n_slides

    @property
    def modes_array(self):
        """Return the spatial modes of each window's fit.

        :return: Modes for each window
        :rtype: numpy.ndarray
        """
        return self._modes_array

    @property
    def amplitudes_array(self):
        """Return the amplitudes of each window's fit.

        :return: amplitudes of each window
        :rtype: numpy.ndarray
        """
        return self._amplitudes_array

    @property
    def omega_array(self):
        """Return the frequencies (omega) of each window's fit.

        :return: omega (a.k.a eigenvalues or time dynamics) for each window
        :rtype: numpy.ndarray
        """
        return self._omega_array

    @property
    def time_array(self):
        """Return the time values contained by each window.

        :return: time values for each fit window
        :rtype: numpy.ndarray
        """
        return self._time_array

    @property
    def window_means_array(self):
        """Return the array of window time means.

        :return: Time mean of the data in each window
        :rtype: numpy.ndarray
        """
        return self._window_means_array

    @property
    def n_components(self):
        """Return the number of frequency bands.

        :return: Number of frequency bands fit in the kmeans clustering
        :rtype: int
        """
        return self._n_components

    @property
    def cluster_centroids(self):
        """Return the frequency band centroids.

        :return: Centroids of the frequency bands
        :rtype: numpy.ndarray
        """
        return self._cluster_centroids

    @property
    def omega_classes(self):
        """Return the frequency band classifications.

        :return: Frequency band classifications, corresponds to omega_array
        :rtype: numpy.ndarray
        """
        return self._omega_classes

    def periods(self):
        """Convert the omega array into periods.

        :return: Time dynamics converted to periods
        :rtype: numpy.ndarray
        """
        if self._omega_array is None:
            raise ValueError("The object must be fit first.")
        frequencies = np.abs(
            self._omega_array[self._omega_classes > 0].imag.flatten()
        )
        return 2 * np.pi / frequencies

    @staticmethod
    def relative_error(x_est, x_true):
        """Helper function for calculating the relative error.

        :param x_est: Estimated values (i.e. from reconstruction)
        :type x_est: numpy.ndarray
        :param x_true: True (or observed) values.
        :type x_true: numpy.ndarray
        :return: Relative error between observations and model.
        :rtype: numpy.ndarray
        """
        return np.linalg.norm(x_est - x_true) / np.linalg.norm(x_true)

    @staticmethod
    def _build_windows(data, window_length, step_size, integer_windows=False):
        """How many times integer slides fit the data for a given step and
        window size.

        :param data: 1D snapshots for fitting
        :type data: numpy.ndarray
        :param window_length: Length of the fitting window in units of time
            steps
        :type window_length: int
        :param step_size:  Distance to slide each window.
        :type step_size: int
        :param integer_windows: Whether to force an integer number of windows
        :type integer_windows: bool
        :return: Number of windows to fit.
        :rtype: int
        """
        if integer_windows:
            n_split = np.floor(data.shape[1] / window_length).astype(int)
        else:
            n_split = data.shape[1] / window_length

        n_steps = int((window_length * n_split))

        # Number of sliding-window iterations
        n_slides = np.floor((n_steps - window_length) / step_size).astype(int)

        return n_slides + 1

    @staticmethod
    def calculate_lv_kern(window_length, corner_sharpness=None):
        """Calculate the kerning window for suppressing real eigenvalues.

        :param corner_sharpness: Parameter specifying how sharp the kerning
            window should be. Default is 16.
        :type corner_sharpness: int
        :param window_length: Size of the window in time steps to kern.
        :type window_length: int
        :return: Kernel for convolving with the windowed data.
        :rtype: np.ndarray
        """

        # Higher = sharper corners
        if corner_sharpness is None:
            corner_sharpness = 16

        lv_kern = (
            np.tanh(
                corner_sharpness
                * np.linspace(0, 1, window_length, endpoint=False)
            )
            - np.tanh(
                corner_sharpness
                * (np.arange(0, window_length) - window_length - 1)
                / window_length
            )
            - 1
        )

        return lv_kern

    @staticmethod
    def build_kern(window_length):
        """Build the convolution kernel for the window reconstruction.

        Each window is convolved with a gaussian filter for the
        reconstruction, which weights points in the middle of the window and
        de-emphasizes the edges of the window that are more poorly fit.

        :param window_length: Length of the data window in units of time
        :type window_length: int
        :return: Gaussian filter of length `window_length`
        :rtype: np.ndarray
        """
        recon_filter_sd = window_length / 8
        recon_filter = np.exp(
            -((np.arange(window_length) - (window_length - 1) / 2) ** 2)
            / recon_filter_sd**2
        )
        return recon_filter

    @staticmethod
    def _data_shape(data):
        """Returns the shape of the data.

        :param data: Data to fit with mrCOSTS.
        :type data: numpy.ndarray
        :return n_time_steps: Number of time steps.
        :rtype n_time_steps: int
        :return n_data_vars: Number of spatial variables.
        :rtype n_data_vars: int
        """
        n_time_steps = np.shape(data)[1]
        n_data_vars = np.shape(data)[0]
        return n_time_steps, n_data_vars

    def _build_proj_basis(self, data, svd_rank=None):
        """Build the projection basis.

        :param data: Data to fit with mrCOSTS.
        :type data: numpy.ndarray
        :param svd_rank: Rank to fit with COSTS.
        :type svd_rank: int
        :return: SVD projection basis for COSTS.
        :rtype: numpy.ndarray
        """
        self._svd_rank = compute_rank(data, svd_rank=svd_rank)
        # Recover the first r modes of the global svd
        return compute_svd(data, svd_rank=svd_rank)[0]

    def _build_initialization(self):
        """Method for making initial guess of DMD eigenvalues.

        :return: First guess of eigenvalues
        :rtype: numpy.ndarray or None
        """
        # User provided initial eigenvalues.
        if self._initialize_artificially and self._init_alpha is not None:
            return self._init_alpha
        # Initial eigenvalue guesses from kmeans clustering.
        elif (
            self._initialize_artificially
            and self._init_alpha is None
            and self._cluster_centroids is not None
        ):
            init_alpha = np.repeat(
                np.sqrt(self._cluster_centroids) * 1j,
                int(self._svd_rank / self._n_components),
            )
            init_alpha = init_alpha * np.tile(
                [1, -1], int(self._svd_rank / self._n_components)
            )
            return init_alpha
        # The user accidentally provided both methods of initializing the
        # eigenvalues.
        if (
            self._initialize_artificially
            and self._init_alpha is not None
            and self._cluster_centroids is not None
        ):
            raise ValueError(
                "Only one of `init_alpha` and `cluster_centroids` can be"
                " provided"
            )

        # In all other cases we return None and let the first iteration of
        # BOPDMD searches for the initial values.
        return None

    def fit(
        self,
        data,
        time,
        window_length,
        step_size,
        verbose=False,
        corner_sharpness=None,
    ):
        """Fit COherent SpatioTemporal Scale separation (COSTS).

        :param data: the input data to decompose (1D snapshots). Dimensions of
            space vs time.
        :type data: numpy.ndarray
        :param time: time series labeling the 1D snapshots
        :type time: numpy.ndarray
        :param window_length: decomposition window length in number of time
            steps.
        :type window_length: int
        :param step_size: Number of time steps to slide forward from the
            previous window.
        :type step_size: int
        :param verbose: notifies progress for fitting. Default is False.
        :type verbose: bool
        :param corner_sharpness: See `calculate_lv_kern`
        :type corner_sharpness: float or int
        """

        # Prepare window and data properties.
        self._window_length = window_length
        self._step_size = step_size
        self._n_time_steps, self._n_data_vars = self._data_shape(data)

        if not self._n_time_steps == time.size:
            raise ValueError("Data and time dimensions do not align.")

        self._n_slides = self._build_windows(
            data,
            self._window_length,
            self._step_size,
        )

        if self._window_length > self._n_time_steps:
            raise ValueError(
                (
                    f"Window length ({self._window_length}) is larger than the "
                    f"time dimension ({self._n_time_steps})"
                )
            )

        # If the window size and step size do not span the data in an integer
        # number of slides, we add one last window that has a smaller step
        # spacing relative to the other window spacings.
        n_slide_last_window = self._n_time_steps - (
            self._step_size * (self._n_slides - 1) + self._window_length
        )
        if n_slide_last_window > 0:
            self._n_slides += 1
            self._non_integer_n_slide = True
        else:
            self._non_integer_n_slide = False

        # Build the projection basis if using a global svd.
        if self._global_svd:
            u = self._build_proj_basis(data, svd_rank=self._svd_rank)
            self._pydmd_kwargs["proj_basis"] = u
            self._pydmd_kwargs["use_proj"] = self._pydmd_kwargs.get(
                "use_proj", False
            )
            self._svd_rank = compute_rank(data, svd_rank=self._svd_rank)
            self._svd_rank_pre_allocate = self._svd_rank
        elif not self._global_svd and self._svd_rank > 0:
            if self._force_even_eigs and self._svd_rank % 2:
                raise ValueError(
                    "svd_rank is odd, but force_even_eigs is True."
                )
            if self._svd_rank > self._n_data_vars:
                raise ValueError(
                    "Rank is larger than the data spatial dimension."
                )
            self._svd_rank_pre_allocate = compute_rank(
                data, svd_rank=self._svd_rank
            )
        # If not using a global svd or a specified svd_rank, local u from
        # each window is used instead. The optimal svd_rank may change when
        # using the locally optimal svd_rank. To deal with this situation in
        # the pre-allocation we give the maximally allowed svd_rank for
        # pre-allocation.
        elif self._max_rank is not None:
            self._svd_rank_pre_allocate = self._max_rank
        else:
            self._svd_rank_pre_allocate = self._n_data_vars

        # Pre-allocate all elements for the sliding window DMD.
        self._time_array = np.zeros((self._n_slides, self._window_length))
        self._modes_array = np.zeros(
            (self._n_slides, self._n_data_vars, self._svd_rank_pre_allocate),
            np.complex128,
        )
        self._omega_array = np.zeros(
            (self._n_slides, self._svd_rank_pre_allocate), np.complex128
        )
        self._amplitudes_array = np.zeros(
            (self._n_slides, self._svd_rank_pre_allocate), np.complex128
        )
        self._window_means_array = np.zeros((self._n_slides, self._n_data_vars))

        # Get initial values for the eigenvalues.
        self._init_alpha = self._build_initialization()

        # Initialize the BOPDMD object.
        optdmd = BOPDMD(
            svd_rank=self._svd_rank,
            init_alpha=self._init_alpha,
            **self._pydmd_kwargs,
        )

        # Round the corners of the windowed data towards zero, which shrinks
        # the real components of the fitted eigenvalues away from unrealistic
        # exponential growth.
        lv_kern = self.calculate_lv_kern(
            self._window_length, corner_sharpness=corner_sharpness
        )

        # Perform the sliding window DMD fitting.
        for k in range(self._n_slides):
            if verbose and k % 50 == 0:
                print(f"{k:} of {self._n_slides:}")

            sample_slice = self.get_window_indices(k)
            data_window = data[:, sample_slice]
            original_time_window = time[:, sample_slice]

            # All windows are fit with the time array reset to start at t=0.
            t_start = original_time_window[:, 0]
            time_window = original_time_window - t_start

            # Subtract off the time mean before rounding corners.
            c = np.mean(data_window, 1, keepdims=True)
            data_window = data_window - c

            # Round the corners of the window.
            data_window = data_window * lv_kern

            # Reset optdmd between iterations
            if not self._global_svd:
                # Get the svd rank for this window. Uses rank truncation when
                # svd_rank is not fixed, i.e. svd_rank = 0, otherwise uses the
                # specified rank.
                _svd_rank = compute_rank(data_window, svd_rank=self._svd_rank)
                # Force svd rank to be even to allow for conjugate pairs.
                if self._force_even_eigs and _svd_rank % 2:
                    _svd_rank += 1
                # Force svd rank to not exceed a user specified amount.
                if self._max_rank is not None:
                    optdmd.svd_rank = min(_svd_rank, self._max_rank)
                else:
                    optdmd.svd_rank = _svd_rank
                optdmd.proj_basis = self._pydmd_kwargs["proj_basis"]

            # Fit the window using the optDMD.
            optdmd.fit(data_window, time_window)

            # Assign the results from this window.
            self._modes_array[k, :, : optdmd.modes.shape[-1]] = optdmd.modes
            self._omega_array[k, : optdmd.eigs.shape[0]] = optdmd.eigs
            self._amplitudes_array[k, : optdmd.eigs.shape[0]] = (
                optdmd.amplitudes
            )
            self._window_means_array[k] = c.flatten()
            self._time_array[k] = original_time_window

            # Reset optdmd between iterations
            if not self._global_svd:
                # The default behavior is to reset the optdmd object to use
                # the initial value from the first window.
                if not self._use_last_freq and not self._reset_alpha_init:
                    optdmd.init_alpha = self._init_alpha
                # Use the eigenvalues from this window to seed the next window.
                elif self._use_last_freq:
                    optdmd.init_alpha = optdmd.eigs
                # Remove the initial guess for the eigenvalues entirely. This
                # is much more computationally expensive.
                elif self._reset_alpha_init:
                    optdmd.init_alpha = None

    def get_window_indices(self, k):
        """Returns the window indices for slide `k`.

        Handles non-integer number of slides by making the last slide
        correspond to `slice(-window_length, None)`.

        :param k: Window to index
        :type k: int
        :return: slice indexing the given window
        :rtype: slice
        """
        # Get the window indices and data.
        sample_start = self._step_size * k
        if k == self._n_slides - 1 and self._non_integer_n_slide:
            return slice(-self._window_length, None)
        return slice(sample_start, sample_start + self._window_length)

    def cluster_omega(
        self,
        n_components,
        kmeans_kwargs=None,
        transform_method=None,
        method=MiniBatchKMeans,
    ):
        """Clusters fitted eigenvalues into frequency bands by the imaginary
        component.

        Assigns the clustering results to the object.

        :param n_components: Hyperparameter for k-means clustering, number of
            clusters.
        :type n_components: int
        :param kmeans_kwargs: Arguments for KMeans clustering. The default is
            random_state = 0.
        :type kmeans_kwargs: dict
        :param transform_method: How to transform omega. See docstring for
            valid options.
        :type transform_method: str or NoneType
        :param method: Clustering method following the sklearn pattern (has
            `fit_predict` and `n_clusters` keywords). Default is
            MiniBatchKMeans.
        :type method: method
        """

        cluster_centroids, omega_classes = self._cluster(
            n_components,
            kmeans_kwargs=kmeans_kwargs,
            transform_method=transform_method,
            method=method,
        )

        # Assign the results to the object.
        self._cluster_centroids = cluster_centroids
        self._omega_classes = omega_classes
        self._transform_method = transform_method
        self._n_components = n_components

    def _cluster(
        self,
        n_components,
        kmeans_kwargs=None,
        transform_method=None,
        method=MiniBatchKMeans,
    ):
        """Clusters fitted eigenvalues into frequency bands by the imaginary
        component.

        Helper function for clustering. Call `cluster_omega` instead.

        :param n_components: Hyperparameter for k-means clustering, number of
            clusters.
        :type n_components: int
        :param kmeans_kwargs: Arguments for KMeans clustering. The default is
            random_state = 0.
        :type kmeans_kwargs: dict or NoneType
        :param transform_method: How to transform omega. See docstring for
            valid options.
        :type transform_method: str or NoneType
        :param method: Clustering method following the sklearn pattern (has
            `fit_predict` and `n_clusters` keywords). Default is
            MiniBatchKMeans.
        :type method: method
        :return omega_classes: Classes defining the frequency bands ordered
            from the largest frequency to the smallest frequency.
        :rtype omega_classes: numpy.ndarray
        :return cluster_centroids: Centroids of the frequency bands. Order
            corresponds to the classes.
        :rtype cluster_centroids: numpy.ndarray
        """
        if kmeans_kwargs is None:
            kmeans_kwargs = {}
            random_state = 0
            kmeans_kwargs["random_state"] = kmeans_kwargs.get(
                "random_state", random_state
            )
        if not hasattr(method, "fit_predict") and callable(
            getattr(method, "fit_predict")
        ):
            raise ValueError(
                "Clustering method must have `fit_predict()` method."
            )
        clustering = method(n_clusters=n_components, **kmeans_kwargs)

        # Reshape the omega array into a 1d array
        omega_rshp = self.omega_array.reshape(
            self._n_slides * self._svd_rank_pre_allocate
        )
        omega_transform = self.transform_omega(
            omega_rshp, transform_method=transform_method
        )

        omega_classes = clustering.fit_predict(np.atleast_2d(omega_transform).T)
        omega_classes = omega_classes.reshape(
            self._n_slides, self._svd_rank_pre_allocate
        )
        cluster_centroids = clustering.cluster_centers_.flatten()

        # Sort the clusters by the centroid magnitude.
        idx = np.argsort(cluster_centroids)
        lut = np.zeros_like(idx)
        lut[idx] = np.arange(n_components)
        omega_classes = lut[omega_classes]
        cluster_centroids = cluster_centroids[idx]

        return cluster_centroids, omega_classes

    def transform_omega(self, omega_array, transform_method="absolute"):
        """Transform omega, primarily for clustering.
        Options for transforming omega are:
            "period": :math:`\\frac{1}{\\omega}`
            "log10": :math:`log10(\\omega)`
            "square_frequencies": :math:`\\omega^2`
            "absolute": :math:`|\\omega|`
        Default value is "absolute". All transformations and clustering are
        performed on the imaginary portion of omega.

        :param omega_array:
        :param transform_method:
        :return: transformed omega array
        :rtype: numpy.ndarray
        """
        # Apply a transformation to omega to improve frequency band separation
        if transform_method == "absolute":
            omega_transform = np.abs(omega_array.imag.astype("float"))
            self._omega_label = r"$|\omega|$"
            self._hist_kwargs = {"bins": 64}
        # Outstanding question: should this be the complex conjugate or
        # the imaginary component squared?
        elif transform_method == "square_frequencies":
            omega_transform = (omega_array.imag**2).real.astype("float")
            self._omega_label = r"$|\omega|^{2}$"
            self._hist_kwargs = {"bins": 64}
        elif transform_method == "log10":
            omega_transform = np.log10(np.abs(omega_array.imag.astype("float")))
            # Impute log10(0) with the smallest non-zero values in log10(omega).
            zero_imputer = omega_transform[np.isfinite(omega_transform)].min()
            omega_transform[~np.isfinite(omega_transform)] = zero_imputer
            self._omega_label = r"$log_{10}(|\omega|)$"
            self._hist_kwargs = {"bins": 64}
        elif transform_method == "period":
            omega_transform = 1 / np.abs(omega_array.imag.astype("float"))
            self._omega_label = "Period"
            self._hist_kwargs = {"bins": 64}
        else:
            raise ValueError(
                f"Transform method {transform_method:} not supported."
            )

        return omega_transform

    def cluster_hyperparameter_sweep(
        self,
        n_components_range=None,
        transform_method=None,
        method=MiniBatchKMeans,
        clustering_kwargs=None,
    ):
        """Hyperparameter search for number of frequency bands.

        Searches for the optimal number of clusters to use in kmeans clustering
        separation of the frequency bands. To best separate frequency bands
        it may be necessary to transform omega. Scores clusters using the
        silhouette score which can be slow to compute.

        Options for transforming omega are:
            "period": :math:`\\frac{1}{\\omega}`
            "log10": :math:`log10(\\omega)`
            "square_frequencies": :math:`\\omega^2`
            "absolute": :math:`|\\omega|`
        Default value is "absolute". All transformations and clustering are
        performed on the imaginary portion of omega.

        :param n_components_range: Range of n_components for the sweep.
        :type n_components_range: numpy.ndarray of ints
        :param transform_method: How to transform the imaginary component of
            omega.
        :type transform_method: str
        :param method: Clustering method following the sklearn pattern (has
            `fit_predict` and `n_clusters` keywords). Default is
            MiniBatchKMeans.
        :param clustering_kwargs: keywords to give to the clustering method.
        :type clustering_kwargs: dict
        :type method: method
        :return: optimal value of `n_components` for clustering.
        """
        if n_components_range is None:
            n_components_range = np.arange(
                np.max((self.svd_rank // 4, 2)),
                self.svd_rank // 2 + 1,
            )
        score = np.zeros_like(n_components_range, float)

        # Reshape the omega array into a 1d array. This is done here and not
        # in the _cluster() helper to reduce the number of times the variable
        # is computed.
        omega_rshp = self.omega_array.reshape(
            self._n_slides * self._svd_rank_pre_allocate
        )
        # Apply the transformation
        omega_transform = self.transform_omega(
            omega_rshp, transform_method=transform_method
        )

        for nind, n in enumerate(n_components_range):
            _, omega_classes = self._cluster(
                n_components=n,
                transform_method=transform_method,
                kmeans_kwargs=clustering_kwargs,
                method=method,
            )

            classes_reshape = omega_classes.reshape(
                self._n_slides * self._svd_rank_pre_allocate
            )

            score[nind] = silhouette_score(
                np.atleast_2d(omega_transform).T,
                np.atleast_2d(classes_reshape).T,
            )

        return n_components_range[np.argmax(score)]

    def plot_omega_histogram(self):
        """Histogram of fit frequencies.

        This plot is useful for assessing if the frequencies bands were well
        separated. A good choice of transformation and clustering will have
        clearly separated clusters.

        :return fig: Figure handle for the plot
        :return ax: Axes handle for the plot
        """

        # Apply the transformation to omega
        omega_transform = self.transform_omega(
            self.omega_array.flatten(), transform_method=self._transform_method
        )

        label = self._omega_label
        hist_kwargs = self._hist_kwargs

        cluster_centroids = self._cluster_centroids

        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        fig, ax = plt.subplots(1, 1)
        ax.hist(omega_transform, **hist_kwargs)
        ax.set_xlabel(label)
        ax.set_ylabel("Count")
        ax.set_title(r"$\omega$ Spectrum & k-Means Centroids")
        [
            ax.axvline(c, color=colors[nc % len(colors)])
            for nc, c in enumerate(cluster_centroids)
        ]

        return fig, ax

    def plot_omega_time_series(self):
        """Time series of transformed omega colored by frequency band.

        :return fig: figure handle for the plot
        :rtype fig: matplotlib.figure()
        :return ax: matplotlib subplot instances
        :rtype ax: matplotlib.Axes()
        """
        fig, ax = plt.subplots(1, 1)
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

        # Apply the transformation to omega
        omega_transform = self.transform_omega(
            self.omega_array.flatten(), transform_method=self._transform_method
        )

        label = self._omega_label

        for ncomponent, component in enumerate(range(self._n_components)):
            ax.plot(
                np.mean(self.time_array, axis=1),
                np.where(
                    self._omega_classes == component,
                    omega_transform.reshape(
                        (self._n_slides, self._svd_rank_pre_allocate)
                    ),
                    np.nan,
                ),
                color=colors[ncomponent % len(colors)],
                ls="None",
                marker=".",
            )
        ax.set_ylabel(label)
        ax.set_xlabel("Time")
        ax.set_title("Time dynamics time series")

        return fig, ax

    def global_reconstruction(self, scale_reconstruction_kwargs=None):
        """Helper function for generating the global reconstruction.

        :param scale_reconstruction_kwargs: Arguments for the scale
            reconstruction.
        :type scale_reconstruction_kwargs: dict
        :return: Global reconstruction (sum of all frequency bands)
        :rtype: numpy.ndarray
        """
        if scale_reconstruction_kwargs is None:
            scale_reconstruction_kwargs = {}
        xr_sep = self.scale_reconstruction(**scale_reconstruction_kwargs)
        x_global_recon = xr_sep.sum(axis=0)
        return x_global_recon

    def scale_reconstruction(
        self,
        include_means=True,
    ):
        """Reconstruct the spatiotemporal features for each frequency band.

        The reconstructed data are convolved with a guassian filter since
        points near the middle of the window are more reliable than points
        at the edge of the window. Note that this will leave the beginning
        and end of time series prone to larger errors. A best practice is
        to cut off `window_length` from each end before further analysis.

        :param include_means: Not API stable
        :return: Reconstruction for each frequency band with dimensions of:
            n_components x n_data_vars x n_time_steps
        :rtype: numpy.ndarray
        """

        # Each individual reconstructed window
        xr_sep = np.zeros(
            (self._n_components, self._n_data_vars, self._n_time_steps)
        )

        # Track the total contribution from all windows to each time step
        xn = np.zeros(self._n_time_steps)

        # Convolve each windowed reconstruction with a gaussian filter.
        # Weights points in the middle of the window and de-emphasizes the
        # edges of the window.
        recon_filter = self.build_kern(self._window_length)

        for k in range(self._n_slides):
            window_indices = self.get_window_indices(k)

            w = self._modes_array[k]
            b = self._amplitudes_array[k]
            omega = copy.deepcopy(np.atleast_2d(self._omega_array[k]).T)
            classification = self._omega_classes[k]

            c = np.atleast_2d(self._window_means_array[k]).T

            # Compute each segment of the reconstructed data starting at "t = 0"
            t = self._time_array[k]
            t_start = t.min()
            t = t - t_start

            xr_sep_window = np.zeros(
                (self._n_components, self._n_data_vars, self._window_length)
            )
            for j in np.unique(self._omega_classes):
                class_index = classification == j
                xr_sep_window[j] = np.linalg.multi_dot(
                    [
                        w[:, class_index],
                        np.diag(b[class_index]),
                        np.exp(omega[class_index] * t),
                    ]
                ).real

                # Add the constant offset to the lowest frequency cluster.
                if include_means and j == np.argmin(self._cluster_centroids):
                    xr_sep_window[j] += c
                xr_sep_window[j] = xr_sep_window[j] * recon_filter

                xr_sep[j, :, window_indices] = (
                    xr_sep[j, :, window_indices] + xr_sep_window[j]
                )

            xn[window_indices] += recon_filter

        xr_sep = xr_sep / xn

        return xr_sep

    def scale_separation(
        self,
        scale_reconstruction_kwargs=None,
    ):
        """Separate the lowest frequency band from the high frequency bands.

        The lowest frequency band should contain the window means and can be
        passed on as the data for the next decomposition level. The high
        frequencies should have frequencies shorter than 1 / window_length.

        :param scale_reconstruction_kwargs: Arguments passed to
            `scale_reconstruction`
        :return xr_low_frequency: Reconstruction of the low frequency
            component with dimensions of n_data_vars x n_time_steps.
        :rtype xr_low_frequency: numpy.ndarray
        :return xr_high_frequency: Sum of all high frequency bands with
            dimensions of n_data_vars x n_time_steps
        :rtype xr_high_frequency: numpy.ndarray
        """

        if scale_reconstruction_kwargs is None:
            scale_reconstruction_kwargs = {}

        xr_sep = self.scale_reconstruction(**scale_reconstruction_kwargs)
        xr_low_frequency = xr_sep[0]
        xr_high_frequency = xr_sep[1:].sum(axis=0)

        return xr_low_frequency, xr_high_frequency

    def plot_scale_separation(
        self,
        data,
        scale_reconstruction_kwargs=None,
        plot_residual=False,
        fig_kwargs=None,
        plot_kwargs=None,
        hf_plot_kwargs=None,
        plot_contours=False,
    ):
        """Plot the scale-separated low and high frequency bands.

        The reconstructions are plotted in a time-space diagram. The high
        frequency component is the sum of all high frequency bands except the
        low frequency band which is plotted separately.

        :param data: Data used for the decomposition. An array of 1D snapshots.
        :type data: numpy.ndarray
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            COSTS fit.
        :type scale_reconstruction_kwargs: dict
        :param plot_residual: If the error should be fit. Will plot
            `data - low frequency - high frequency` yielding the error in
            absolute units.
        :type plot_residual: bool
        :param fig_kwargs: Arguments for the figure creation.
        :type fig_kwargs: dict
        :param plot_kwargs: Arguments for plotting the low frequency and data.
        :type plot_kwargs: dict
        :param hf_plot_kwargs: Arguments for the high frequency plotting.
        :type hf_plot_kwargs: dict
        :param plot_contours: Indicates if contours should be plotted.
        :type plot_contours: bool
        :return fig: figure handle for the plot
        :rtype fig: matplotlib.figure()
        :return axes: matplotlib subplot instances
        :rtype fig: matplotlib.Axes()
        """
        if scale_reconstruction_kwargs is None:
            scale_reconstruction_kwargs = {}

        xr_low_frequency, xr_high_frequency = self.scale_separation(
            scale_reconstruction_kwargs
        )

        if fig_kwargs is None:
            fig_kwargs = {}
        fig_kwargs["sharex"] = fig_kwargs.get("sharex", True)
        fig_kwargs["figsize"] = fig_kwargs.get("figsize", (6, 4))

        if plot_kwargs is None:
            plot_kwargs = {}
        plot_kwargs["vmin"] = plot_kwargs.get("vmin", -np.abs(data).max())
        plot_kwargs["vmax"] = plot_kwargs.get("vmax", np.abs(data).max())
        plot_kwargs["cmap"] = plot_kwargs.get("cmap", "cividis")

        if hf_plot_kwargs is None:
            hf_plot_kwargs = {}
        hf_plot_kwargs["vmin"] = hf_plot_kwargs.get(
            "vmin", -np.abs(xr_high_frequency).max()
        )
        hf_plot_kwargs["vmax"] = hf_plot_kwargs.get(
            "vmax", np.abs(xr_high_frequency).max()
        )
        hf_plot_kwargs["cmap"] = hf_plot_kwargs.get("cmap", "RdBu_r")

        if plot_residual:
            fig, axes = plt.subplots(4, 1, **fig_kwargs)
        else:
            fig, axes = plt.subplots(3, 1, **fig_kwargs)

        ax = axes[0]
        ax.pcolormesh(data, **plot_kwargs)
        if plot_contours:
            ax.contour(data, colors=["k"])
        ax.set_title(
            (
                f"Input data at decomposition window length ="
                f" {self._window_length:}"
            )
        )
        ax.set_ylabel("Space (-)")

        ax = axes[1]
        ax.set_title("Reconstruction, low frequency")
        ax.pcolormesh(xr_low_frequency, **plot_kwargs)
        if plot_contours:
            ax.contour(data, colors=["k"])
        ax.set_ylabel("Space (-)")

        ax = axes[2]
        ax.set_title("Reconstruction, high frequency")
        ax.pcolormesh(xr_high_frequency, **hf_plot_kwargs)
        ax.set_ylabel("Space (-)")

        if plot_residual:
            ax = axes[3]
            ax.set_title("Residual")
            ax.pcolormesh(
                data - xr_high_frequency - xr_low_frequency, **hf_plot_kwargs
            )
            ax.set_ylabel("Space (-)")

        axes[-1].set_xlabel("Time (-)")
        fig.tight_layout()

        return fig, axes

    def plot_reconstructions(
        self,
        data,
        plot_period=False,
        scale_reconstruction_kwargs=None,
        plot_residual=False,
        fig_kwargs=None,
        plot_kwargs=None,
        hf_plot_kwargs=None,
        plot_contours=False,
    ):
        """Time-space plots for each individual frequency band and the fitted
        data.

        :param data: Data used for the decomposition. An array of 1D snapshots.
        :type data: numpy.ndarray
        :param plot_period:
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            COSTS fit.
        :type scale_reconstruction_kwargs: dict
        :param plot_residual: Indicates if the residual of the fit should be
            plotted
        :type plot_residual: bool
        :param fig_kwargs: Arguments for the figure creation.
        :type fig_kwargs: dict
        :param plot_kwargs: Arguments for plotting the low frequency and data.
        :type plot_kwargs: dict
        :param hf_plot_kwargs: Arguments for the high frequency plotting.
        :type hf_plot_kwargs: dict
        :param plot_contours: Indicates if contours should be plotted.
        :type plot_contours: bool
        :return fig: figure handle for the plot
        :rtype fig: matplotlib.figure()
        :return axes: matplotlib subplot instances
        :rtype axes: matplotlib.Axes()
        """
        if scale_reconstruction_kwargs is None:
            scale_reconstruction_kwargs = {}

        xr_sep = self.scale_reconstruction(**scale_reconstruction_kwargs)

        if fig_kwargs is None:
            fig_kwargs = {}
        fig_kwargs["sharex"] = fig_kwargs.get("sharex", True)
        fig_kwargs["figsize"] = fig_kwargs.get(
            "figsize", (6, 1.5 * len(self._cluster_centroids) + 1)
        )

        # Low frequency and input data often require separate plotting
        # parameters.
        if plot_kwargs is None:
            plot_kwargs = {}
        plot_kwargs["vmin"] = plot_kwargs.get("vmin", -np.abs(data).max())
        plot_kwargs["vmax"] = plot_kwargs.get("vmax", np.abs(data).max())
        plot_kwargs["cmap"] = plot_kwargs.get("cmap", "cividis")

        # High frequency components often require separate plotting parameters.
        if hf_plot_kwargs is None:
            hf_plot_kwargs = {}
        hf_plot_kwargs["vmin"] = hf_plot_kwargs.get(
            "vmin", -np.abs(xr_sep[1:, :, :]).max()
        )
        hf_plot_kwargs["vmax"] = hf_plot_kwargs.get(
            "vmax", np.abs(xr_sep[1:, :, :]).max()
        )
        hf_plot_kwargs["cmap"] = hf_plot_kwargs.get("cmap", "RdBu_r")

        # Determine the number of plotting elements, which changes depending on
        # if the residual is included.
        if plot_residual:
            num_plot_elements = len(self._cluster_centroids) + 2
        else:
            num_plot_elements = len(self._cluster_centroids) + 1
        fig, axes = plt.subplots(
            num_plot_elements,
            1,
            **fig_kwargs,
        )

        ax = axes[0]
        ax.pcolormesh(data.real, **plot_kwargs)
        if plot_contours:
            ax.contour(data.real, colors=["k"])
        ax.set_ylabel("Space (-)")
        ax.set_xlabel("Time (-)")
        ax.set_title(
            f"Input Data at decomposition window length = {self._window_length}"
        )
        for n_cluster, cluster in enumerate(self._cluster_centroids):
            if plot_period:
                x = 2 * np.pi / cluster
                title = "Reconstruction, central period={:.2f}"
            else:
                x = cluster
                title = "Reconstruction, central eig={:.2f}"

            ax = axes[n_cluster + 1]
            xr_scale = xr_sep[n_cluster, :, :]
            if n_cluster == 0:
                ax.pcolormesh(xr_scale, **plot_kwargs)
                if plot_contours:
                    ax.contour(xr_scale, colors=["k"])
            else:
                ax.pcolormesh(xr_scale, **hf_plot_kwargs)
            ax.set_ylabel("Space (-)")
            ax.set_title(title.format(x))

        if plot_residual:
            ax = axes[-1]
            ax.set_title("Residual")
            ax.pcolormesh(data - xr_sep.sum(axis=0), **hf_plot_kwargs)
            ax.set_ylabel("Space (-)")

        axes[-1].set_xlabel("Time (-)")
        fig.tight_layout()

        return fig, axes

    def plot_error(
        self, data, scale_reconstruction_kwargs=None, plot_kwargs=None
    ):
        """Plots the error for the COSTS fit

        Plots are a space-time diagram assuming a 1D spatial dimension.

        Determining the error requires the input data and the fit will be
        reconstructed.

        :param data: Data on which COSTS was fit
        :type data: numpy.ndarray
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            fit.
        :type scale_reconstruction_kwargs: dict
        :param plot_kwargs: Arguments passed to costs.plot_error().
        :type scale_reconstruction_kwargs: dict
        :return fig: figure handle for the plot
        :rtype fig: matplotlib.figure()
        :return axes: matplotlib subplot instances
        :rtype axes: matplotlib.Axes()
        """
        if scale_reconstruction_kwargs is None:
            scale_reconstruction_kwargs = {}
        if plot_kwargs is None:
            plot_kwargs = {}
        plot_kwargs["vmin"] = plot_kwargs.get("vmin", -3)
        plot_kwargs["vmax"] = plot_kwargs.get("vmax", 3)
        plot_kwargs["cmap"] = plot_kwargs.get("cmap", "RdBu_r")

        global_reconstruction = self.global_reconstruction(
            scale_reconstruction_kwargs=scale_reconstruction_kwargs
        )

        fig_glbl_r, ax_glbl_r = plt.subplots(
            1,
            1,
        )
        im = ax_glbl_r.pcolormesh(
            (global_reconstruction.real - data.real) / data.real * 100,
            **plot_kwargs,
        )

        cbar = fig_glbl_r.colorbar(im)
        cbar.set_label("% error")

        ax_glbl_r.set_xlabel("time (-)")
        ax_glbl_r.set_ylabel("space (-)")
        re = self.relative_error(global_reconstruction.real, data)
        ax_glbl_r.set_title(f"Error in Global Reconstruction = {re:.2}")

    def plot_time_series(
        self,
        space_index,
        data,
        scale_reconstruction_kwargs=None,
        include_residual=False,
    ):
        """Plots CoSTS for a single spatial point.

        Includes the input data for decomposition, the low-frequency component
        for the next decomposition level, the residual of the high frequency
        component, and the reconstructions of the frequency bands for the point.

        :param space_index: Index of the point in space for the 1D snapshot.
        :type space_index: int
        :param data: Original data, only necessary for level=0.
        :type data: numpy.ndarray
        :param scale_reconstruction_kwargs: Arguments for reconstructing the
            fit.
        :type scale_reconstruction_kwargs: dict
        :param include_residual:
        :return fig: figure handle for the plot
        :rtype fig: matplotlib.figure()
        :return axes: matplotlib subplot instances
        :rtype axes: matplotlib.Axes()
        """
        ground_truth_mean = data.mean(axis=1)
        ground_truth = (data.T - ground_truth_mean).T
        ground_truth = ground_truth[space_index, :]
        ground_truth_mean = ground_truth_mean[space_index]

        if scale_reconstruction_kwargs is None:
            scale_reconstruction_kwargs = {}
        xr_sep = self.scale_reconstruction(**scale_reconstruction_kwargs)

        fig, axes = plt.subplots(
            nrows=self.n_components + 2,
            sharex=True,
            sharey=True,
            figsize=(8, np.max((8, 1.5 * self.n_components))),
        )

        ax = axes[0]
        ax.plot(ground_truth, color="k")
        ax.plot(
            xr_sep.sum(axis=0)[space_index, :] - ground_truth_mean,
            color="r",
            lw=0.5,
        )
        ax.set_title(
            f"window={self._window_length}, black=input data, "
            f"red=reconstruction"
        )
        ax.set_ylabel("Amp.")
        ax.set_xlabel("Time")

        ax = axes[1]
        ax.plot(
            ground_truth - (xr_sep[1:, :, :].sum(axis=0))[space_index, :],
            color="k",
        )

        for n in range(self.n_components):
            ax = axes[n + 1]
            if n == 0:
                title = (
                    "blue = Low-frequency component, black = high "
                    "frequency residual"
                )
                ax.plot(xr_sep[n, space_index, :] - ground_truth_mean)
            else:
                period = 2 * np.pi / self.cluster_centroids[n]
                title = f"Band period = {period:.0f} window length"
                ax.plot(xr_sep[n, space_index, :])
            ax.set_title(title)
            ax.set_ylabel("Amp.")
            ax.set_xlabel("Time")

        ax = axes[-1]
        ax.plot(ground_truth, color="k", label="Smoothed data")
        ax.plot(
            (xr_sep[1:, :, :].sum(axis=0))[space_index, :],
            label="High-frequency",
        )
        ax.plot(
            xr_sep[0, space_index, :] - ground_truth_mean, label="Low-frequency"
        )
        if include_residual:
            ax.plot(
                ground_truth - xr_sep.sum(axis=0)[space_index, :],
                label="Residual",
            )
            ax.set_title(
                (
                    "black=input data, yellow=low-frequency, "
                    "blue=high-frequency, red=residual"
                )
            )
        else:
            ax.set_title(
                "black=input data, yellow=low-frequency, blue=high-frequency"
            )

        ax.set_ylabel("Amp.")
        ax.set_xlabel("Time")
        ax.set_xlim(0, self._n_time_steps)
        fig.tight_layout()

        return fig, axes

    def to_xarray(self):
        """Build an xarray dataset from the fitted CoSTS object.

        The CoSTS object is converted to an xarray dataset, which allows
        saving the computationally expensive results, e.g., between iterations.

        The reconstructed data are not included since their size can rapidly
        explode to unexpected sizes. e.g., a 30MB dataset, decomposed at 6
        levels with an average number of frequency bands across decomposition
        levels equal to 8 becomes 1.3GB once reconstructed for each band.

        The functions `to_xarray` and `from_xarray` should allow for a complete
        round trip of the COSTS object without alteration.

        :return: COSTS fit in xarray format
        :rtype: xarray.Dataset
        """
        ds = xr.Dataset(
            {
                "omega": (("window_time_means", "svd_rank"), self.omega_array),
                "omega_classes": (
                    ("window_time_means", "svd_rank"),
                    self.omega_classes,
                ),
                "amplitudes": (
                    ("window_time_means", "svd_rank"),
                    self.amplitudes_array,
                ),
                "modes": (
                    ("window_time_means", "space", "svd_rank"),
                    self.modes_array,
                ),
                "window_means": (
                    ("window_time_means", "space"),
                    self.window_means_array,
                ),
                "cluster_centroids": (
                    "frequency_band",
                    self._cluster_centroids,
                ),
            },
            coords={
                "window_time_means": np.mean(self.time_array, axis=1),
                "slide": ("window_time_means", np.arange(self._n_slides)),
                "svd_rank": np.arange(self.svd_rank),
                "space": np.arange(self._n_data_vars),
                "frequency_band": np.arange(self.n_components),
                "window_index": np.arange(self._window_length),
                "time": (
                    ("window_time_means", "window_index"),
                    self.time_array,
                ),
            },
            attrs={
                "svd_rank": self.svd_rank,
                "omega_transformation": self._xarray_sanitize(
                    self._transform_method
                ),
                "n_slides": self._n_slides,
                "window_length": self._window_length,
                "num_frequency_bands": self.n_components,
                "n_data_vars": self._n_data_vars,
                "n_time_steps": self._n_time_steps,
                "step_size": self._step_size,
                "non_integer_n_slide": self._non_integer_n_slide,
                "global_svd": self._global_svd,
            },
        )

        for kw, kw_val in self._pydmd_kwargs.items():
            ds.attrs[f"pydmd_kwargs__{kw}"] = self._xarray_sanitize(kw_val)

        return ds

    def from_xarray(self, ds):
        """Convert xarray Dataset into a fitted CoSTS object.

        The functions `to_xarray` and `from_xarray` should allow for a complete
        round trip of the COSTS object without alteration.

        :return: Previously fitted COSTS object.
        """

        self._omega_array = ds.omega.values
        self._omega_classes = ds.omega_classes.values
        self._amplitudes_array = ds.amplitudes.values
        self._modes_array = ds.modes.values
        self._window_means_array = ds.window_means.values
        self._cluster_centroids = ds.cluster_centroids.values
        self._time_array = ds.time.values
        self._n_slides = ds.attrs["n_slides"]
        self._svd_rank = ds.attrs["svd_rank"]
        self._n_data_vars = ds.attrs["n_data_vars"]
        self._n_time_steps = ds.attrs["n_time_steps"]
        self._n_components = ds.attrs["num_frequency_bands"]
        self._non_integer_n_slide = ds.attrs["non_integer_n_slide"]
        self._step_size = ds.attrs["step_size"]
        self._window_length = ds.attrs["window_length"]
        self._global_svd = ds.attrs["global_svd"]

        self._pydmd_kwargs = {}
        for attr in ds.attrs:
            if "pydmd_kwargs" in attr:
                new_attr_name = attr.replace("pydmd_kwargs__", "")
                self._pydmd_kwargs[new_attr_name] = self._xarray_unsanitize(
                    ds.attrs[attr]
                )
                if new_attr_name == "eig_constraints":
                    self._pydmd_kwargs[new_attr_name] = set(
                        self._pydmd_kwargs[new_attr_name]
                    )
            elif "omega_transformation" in attr:
                self._transform_method = self._xarray_unsanitize(ds.attrs[attr])

        return self

    @staticmethod
    def _xarray_sanitize(value):
        """Handle Nones in the pydmd_kwargs (i.e., used default values)

        Netcdf cannot handle NoneTypes. To allow the xarray DataSet to be
        saved to file we have to "sanitize" the NoneTypes. These two functions
        allow for a round trip recovery of `pydmd_kwargs`.

        :param value: Value to be stored in the attributes of an xarray Dataset.
        :return: value unaltered except if value NoneType.
        """
        if value is None:
            value = "None"
        elif isinstance(value, set):
            value = list(value)
        elif callable(value):
            value = f"Custom function {value.__name__}"
        return value

    @staticmethod
    def _xarray_unsanitize(value):
        """Handle Nones in the pydmd_kwargs (i.e., used default values)

        Netcdf cannot handle NoneTypes. To allow the xarray DataSet to be
        saved to file we have to "sanitize" the NoneTypes. These two functions
        allow for a round trip recovery of `pydmd_kwargs`.

        :param value: Value stored in the attributes of an xarray Dataset.
        :return: value unaltered except if value is the string "None".
        """
        # To handle the varying behavior between python versions when evaluating
        # a mixed type statement we have to try to catch cases when `value` is
        # an array (e.g., `proj_basis` kwarg). The except block should not be
        # triggered but is meant to catch edge cases when the user provides
        # an unexpected type (e.g., tuple).
        if not hasattr(value, "shape"):
            try:
                if value == "None":
                    return None
            except ValueError:
                return value
        return value



"""
Hankel pre-processing.
"""

from __future__ import annotations

import sys
from typing import Tuple, Union

import numpy as np

from pydmd.dmdbase import DMDBase
from pydmd.preprocessing.pre_post_processing import (
    PrePostProcessing,
    PrePostProcessingDMD,
)
from pydmd.utils import pseudo_hankel_matrix

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

_ReconstructionMethodType = Union[str, np.ndarray, list, tuple]


def hankel_preprocessing(
    dmd: DMDBase,
    d: int,
    reconstruction_method: _ReconstructionMethodType = "first",
):
    """
    Hankel pre-processing.

    :param dmd: DMD instance to be wrapped.
    :param d: Hankel matrix rank
    :param reconstruction_method: Reconstruction method.
    """
    if isinstance(reconstruction_method, (tuple, list)):
        reconstruction_method = np.asarray(reconstruction_method)

    pre_post_processing = _HankelPrePostProcessing(
        d=d, reconstruction_method=reconstruction_method
    )
    return PrePostProcessingDMD(dmd, pre_post_processing)


def _reconstructions(rec: np.ndarray, d: int):
    """
    Build a collection of all the available versions of the given
    `timeindex`. The indexing of time instants is the same used for
    :func:`reconstructed_data`. For each time instant there are at least
    one and at most `d` versions. If `timeindex` is `None` the function
    returns the whole collection, for all the time instants.

    :param rec: reconstructed data.
    :param d: Hankel matrix rank.
    :param timeindex: The index of the time snapshot.
    :return: a collection of all the available versions for the given
        time snapshot, or for all the time snapshots if `timeindex` is
        `None` (in the second case, time varies along the first dimension
        of the array returned).
    """
    space_dim = rec.shape[-2] // d
    time_instants = rec.shape[-1] + d - 1

    rec_snapshots_shape = (time_instants, d, space_dim)
    reconstructed_snapshots = np.full(
        rec_snapshots_shape, np.nan, dtype=rec.dtype
    )
    time_idxes = np.arange(d)[None].repeat(rec.shape[-1], axis=0)
    time_idxes += np.arange(rec.shape[-1])[:, None]
    d_idxes = np.arange(d)[None].repeat(rec.shape[-1], axis=0)
    splitted = np.array(np.split(rec, d, axis=-2))
    splitted = np.moveaxis(splitted, -1, -3)

    reconstructed_snapshots[..., time_idxes, d_idxes, :] = splitted
    return reconstructed_snapshots


def _first_reconstructions(reconstructions: np.ndarray, d: int) -> np.ndarray:
    """Return the first occurrence of each snapshot available in the given
    matrix (which must be the result of `self._sub_dmd.reconstructed_data`,
    or have the same shape).

    :param reconstructions: A matrix of (higher-order) snapshots having
        shape `(space*self.d, time_instants)`.
    :param d: Hankel matrix rank.
    :return: The first snapshot that occurs in `reconstructions` for each
        available time instant.
    """
    n_time = reconstructions.shape[-3]
    time_idxes = np.arange(n_time)
    d_idxes = np.arange(n_time)
    d_idxes[d - 1 :] = d - 1
    return reconstructions[..., time_idxes, d_idxes, :].swapaxes(-1, -2)


class _HankelPrePostProcessing(PrePostProcessing):
    def __init__(
        self, *, d: int, reconstruction_method: _ReconstructionMethodType
    ):
        self._d = d
        self._reconstruction_method = reconstruction_method

    @override
    def pre_processing(self, X: np.ndarray) -> Tuple[None, np.ndarray]:
        return None, pseudo_hankel_matrix(X, self._d)

    @override
    def post_processing(
        self, pre_processing_output: None, Y: np.ndarray
    ) -> np.ndarray:
        rec = _reconstructions(Y, d=self._d)
        rec = np.ma.array(rec, mask=np.isnan(rec))

        if isinstance(self._reconstruction_method, str):
            if self._reconstruction_method == "first":
                result = _first_reconstructions(rec, d=self._d)
            elif self._reconstruction_method == "mean":
                result = np.nanmean(rec, axis=1).T
        elif isinstance(self._reconstruction_method, np.ndarray):
            result = np.ma.average(
                rec, axis=1, weights=self._reconstruction_method
            ).T
        else:
            raise ValueError(f"{self._reconstruction_method=} not recognized")

        return result.filled(fill_value=0)



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises

from pydmd.pidmd import PiDMD


def error(true, comp):
    """Helper function that computes and returns relative error."""
    return np.linalg.norm(comp - true) / np.linalg.norm(true)


# Use random matrix as dummy data.
rng = np.random.default_rng(seed=42)
X = rng.standard_normal((25, 1000))
X1 = X[:, :-1]
X2 = X[:, 1:]

# Compute A via pseudoinverse as benchmark.
A_dmd = X2.dot(np.linalg.pinv(X1))
dmd_error = error(X2, A_dmd.dot(X1))
error_tol = 0.45 * dmd_error


def assert_accurate(A):
    """
    Helper method that, given a computed A matrix, tests that A is an
    accurate operator relative to that obtained via standard DMD.
    """
    computed_error = error(X2, A.dot(X1))
    assert_allclose(computed_error, dmd_error, atol=error_tol)


def assert_all_zero(A):
    """
    Helper method that, given a matrix A, tests that A is
    approximately a matrix of only zero entries.
    """
    assert_allclose(np.linalg.norm(A), 0, atol=1e-14)


def assert_circulant(A):
    """
    Helper method that, given a matrix A, tests that A is a circulant matrix.
    """
    for i in range(1, len(A)):
        assert_allclose(np.roll(A[i, :], -i), A[0, :], atol=1e-14)


def assert_block_circulant(A):
    """
    Helper method that, given a matrix A, tests that A is block circulant.
    Assumes A is a square matrix composed of equally-sized square blocks.
    """
    block_n = int(np.sqrt(len(A)))
    for i in range(1, block_n):
        block_row_i = A[i * block_n : (i + 1) * block_n, :]
        rolled_block_row_i = np.roll(block_row_i, -i * block_n, axis=1)
        assert_allclose(rolled_block_row_i, A[:block_n, :], atol=1e-14)


def test_invalid_manifold():
    # Test that an error is thrown if an invalid manifold is given.
    with raises(ValueError):
        PiDMD("some_manifold").fit(X)


def test_unitary():
    pidmd = PiDMD("unitary", compute_A=True).fit(X)
    # Ensure that the A matrix is unitary and accurate.
    assert_allclose(pidmd.A.conj().T.dot(pidmd.A), np.eye(25), atol=1e-14)
    assert_accurate(pidmd.A)


def test_uppertriangular():
    # Throw error if model is fitted without computing A.
    with raises(ValueError):
        pidmd = PiDMD("uppertriangular").fit(X)
    pidmd = PiDMD("uppertriangular", compute_A=True).fit(X)
    # Ensure that the A matrix is uppertriangular and accurate.
    assert_all_zero(np.tril(pidmd.A, k=-1))
    assert_accurate(pidmd.A)


def test_lowertriangular():
    # Throw error if model is fitted without computing A.
    with raises(ValueError):
        pidmd = PiDMD("lowertriangular").fit(X)
    pidmd = PiDMD("lowertriangular", compute_A=True).fit(X)
    # Ensure that the A matrix is lowertriangular and accurate.
    assert_all_zero(np.triu(pidmd.A, k=1))
    assert_accurate(pidmd.A)


def test_diagonal():
    # Throw error if manifold_opt is formatted incorrectly.
    with raises(ValueError):
        pidmd = PiDMD("diagonal", manifold_opt=-1, compute_A=True).fit(X)

    # Test diagonal model with integer manifold_opt.
    pidmd = PiDMD("diagonal", manifold_opt=1, compute_A=True).fit(X)
    assert_all_zero(pidmd.A - np.diag(np.diag(pidmd.A)))
    assert_accurate(pidmd.A)

    # Test tridiagonal model with tuple manifold_opt.
    pidmd = PiDMD("diagonal", manifold_opt=(2, 2), compute_A=True).fit(X)
    assert_all_zero(np.triu(pidmd.A, k=2) + np.tril(pidmd.A, k=-2))
    assert_accurate(pidmd.A)

    # Test model with custom (2, nx) manifold_opt.
    custom_opt = np.hstack((np.ones((25, 1)), 4 * np.ones((25, 1))))
    pidmd = PiDMD("diagonal", manifold_opt=custom_opt, compute_A=True).fit(X)
    assert_all_zero(np.triu(pidmd.A, k=4) + np.tril(pidmd.A, k=-1))
    assert_accurate(pidmd.A)


def test_symmetric():
    pidmd = PiDMD("symmetric", compute_A=True).fit(X)
    # Ensure that the A matrix is symmetric and accurate.
    assert_allclose(pidmd.A, pidmd.A.T)
    assert_accurate(pidmd.A)


def test_skewsymmetric():
    pidmd = PiDMD("skewsymmetric", compute_A=True).fit(X)
    # Ensure that the A matrix is skewsymmetric and accurate.
    assert_allclose(pidmd.A, -pidmd.A.T, atol=1e-14)
    assert_accurate(pidmd.A)


def test_toeplitz():
    # Throw error if model is fitted without computing A.
    with raises(ValueError):
        pidmd = PiDMD("toeplitz").fit(X)
    pidmd = PiDMD("toeplitz", compute_A=True).fit(X)
    # Ensure that the A matrix is toeplitz and accurate.
    for k in range(-24, 25):
        k_diag = np.diag(pidmd.A, k=k)
        assert_allclose(k_diag, k_diag[0] * np.ones(k_diag.shape))
    assert_accurate(pidmd.A)


def test_hankel():
    # Throw error if model is fitted without computing A.
    with raises(ValueError):
        pidmd = PiDMD("hankel").fit(X)
    pidmd = PiDMD("hankel", compute_A=True).fit(X)
    # Ensure that the A matrix is hankel and accurate.
    for k in range(-24, 25):
        k_diag = np.diag(np.fliplr(pidmd.A), k=k)
        assert_allclose(k_diag, k_diag[0] * np.ones(k_diag.shape))
    assert_accurate(pidmd.A)


def test_circulant():
    # Test circulant case.
    pidmd = PiDMD("circulant", compute_A=True).fit(X)
    assert_circulant(pidmd.A)
    assert_accurate(pidmd.A)

    # Test circulant_unitary case.
    pidmd = PiDMD("circulant_unitary", compute_A=True).fit(X)
    assert_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    assert_allclose(pidmd.A.conj().T.dot(pidmd.A), np.eye(25), atol=1e-14)

    # Test circulant_symmetric case.
    pidmd = PiDMD("circulant_symmetric", compute_A=True).fit(X)
    assert_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    assert_allclose(pidmd.A, pidmd.A.T)

    # Test circulant_skewsymmetric case.
    pidmd = PiDMD("circulant_skewsymmetric", compute_A=True).fit(X)
    assert_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    assert_allclose(pidmd.A, -pidmd.A.T, atol=1e-14)


def test_symmetric_tridiagonal():
    pidmd = PiDMD("symmetric_tridiagonal", compute_A=True).fit(X)
    # Ensure that the A matrix is symmetric, tridiagonal, and accurate.
    assert_allclose(pidmd.A, pidmd.A.T)
    assert_all_zero(np.triu(pidmd.A, k=2) + np.tril(pidmd.A, k=-2))
    assert_accurate(pidmd.A)


def test_BC():
    # Throw error if model is fitted without computing A.
    with raises(ValueError):
        pidmd = PiDMD("BC", manifold_opt=(5, 5)).fit(X)

    # Throw error if model is fitted without providing block size.
    with raises(ValueError):
        pidmd = PiDMD("BC", compute_A=True).fit(X)

    # BC (block circulant)
    pidmd = PiDMD("BC", manifold_opt=(5, 5), compute_A=True).fit(X)
    assert_block_circulant(pidmd.A)
    assert_accurate(pidmd.A)

    # BCTB (block circulant with tridiagonal blocks)
    pidmd = PiDMD("BCTB", manifold_opt=(5, 5), compute_A=True).fit(X)
    assert_block_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    for i in range(5):
        for j in range(5):
            block = pidmd.A[i * 5 : (i + 1) * 5, j * 5 : (j + 1) * 5]
            assert_all_zero(np.triu(block, k=2) + np.tril(block, k=-2))

    # BCCB (block circulant with circulant blocks)
    pidmd = PiDMD("BCCB", manifold_opt=(5, 5), compute_A=True).fit(X)
    assert_block_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    for i in range(5):
        for j in range(5):
            block = pidmd.A[i * 5 : (i + 1) * 5, j * 5 : (j + 1) * 5]
            assert_circulant(block)

    # BCCB and unitary
    pidmd = PiDMD("BCCBunitary", manifold_opt=(5, 5), compute_A=True).fit(X)
    assert_block_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    for i in range(5):
        for j in range(5):
            block = pidmd.A[i * 5 : (i + 1) * 5, j * 5 : (j + 1) * 5]
            assert_circulant(block)
    assert_allclose(pidmd.A.conj().T.dot(pidmd.A), np.eye(25), atol=1e-14)

    # BCCB and symmetric
    pidmd = PiDMD("BCCBsymmetric", manifold_opt=(5, 5), compute_A=True).fit(X)
    assert_block_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    for i in range(5):
        for j in range(5):
            block = pidmd.A[i * 5 : (i + 1) * 5, j * 5 : (j + 1) * 5]
            assert_circulant(block)
    assert_allclose(pidmd.A, pidmd.A.T)

    # BCCB and skewsymmetric
    pidmd = PiDMD("BCCBskewsymmetric", manifold_opt=(5, 5), compute_A=True).fit(
        X
    )
    assert_block_circulant(pidmd.A)
    assert_accurate(pidmd.A)
    for i in range(5):
        for j in range(5):
            block = pidmd.A[i * 5 : (i + 1) * 5, j * 5 : (j + 1) * 5]
            assert_circulant(block)
    assert_allclose(pidmd.A, -pidmd.A.T, atol=1e-14)



import numpy as np
from numpy.testing import assert_allclose
from pytest import raises

from pydmd.edmd import EDMD

# Simulate testing data using 500 random initial conditions.
J = np.array([[0.9, -0.1], [0.0, 0.8]])
rng = np.random.default_rng(seed=42)
X = rng.standard_normal((2, 500))
Y = J.dot(X)

# Additionally use model to propagate a single initial condition forward.
X2 = np.empty(X.shape)
X2[:, 0] = X[:, 0]
for i in range(X.shape[1] - 1):
    X2[:, i + 1] = J.dot(X2[:, i])

# Compute the first 8 ground truth eigenvalues and eigenfunctions along
# the defined xy-grid (excluding the eigenfunction, eigenvalue pair 1,1).
x_vals = np.linspace(-5, 5, 11)
y_vals = np.linspace(5, -5, 11)
X_grid, Y_grid = np.meshgrid(x_vals, y_vals)
eigenfunctions_true = np.empty((8, *X_grid.shape))
eigenvalues_true = np.empty(8)
for idx, (i, j) in enumerate(
    zip([1, 2, 0, 3, 1, 4, 2, 0], [0, 0, 1, 0, 1, 0, 1, 2])
):
    eigenfunc = np.multiply(((X_grid - Y_grid) / np.sqrt(2)) ** i, Y_grid**j)
    eigenfunc /= np.linalg.norm(eigenfunc, np.inf)
    eigenfunctions_true[idx] = eigenfunc
    eigenvalues_true[idx] = (0.9**i) * (0.8**j)


def relative_error(mat, mat_true):
    """
    Computes and returns the relative error between two matrices.
    """
    return np.linalg.norm(mat - mat_true) / np.linalg.norm(mat_true)


def test_eigs():
    """
    Test that EDMD can sucessfully recover the true eigenvalues
    of the toy system from the original EDMD paper.
    """
    edmd = EDMD(
        svd_rank=15,
        kernel_metric="poly",
        kernel_params={"gamma": 1, "coef0": 1, "degree": 4},
    )
    edmd.fit(X, Y)
    sorted_inds = np.argsort(-np.abs(edmd.eigs))
    edmd_eigs = edmd.eigs[sorted_inds][1:9]
    assert_allclose(edmd_eigs, eigenvalues_true)


def test_eigenfunctions():
    """
    Test that EDMD can sucessfully recover the true eigenfunctions
    of the toy system from the original EDMD paper.
    """
    edmd = EDMD(
        svd_rank=15,
        kernel_metric="poly",
        kernel_params={"gamma": 1, "coef0": 1, "degree": 4},
    )
    edmd.fit(X, Y)

    # Evaluate eigenfunctions from EDMD along the grid.
    eigenfunctions = np.empty((15, *X_grid.shape))
    for y_idx, y in enumerate(Y_grid[:, 0]):
        for x_idx, x in enumerate(X_grid[0, :]):
            xy = np.array([x, y])
            eigenfunctions[:, y_idx, x_idx] = edmd.eigenfunctions(xy).real

    # Scale eigenfunctions to have infinity norm 1.
    for eigenfunction in eigenfunctions:
        eigenfunction /= np.linalg.norm(eigenfunction, np.inf)

    # Sort the eigenfunctions based on eigenvalue magnitude.
    sorted_inds = np.argsort(-np.abs(edmd.eigs))
    edmd_eigenfunctions = eigenfunctions[sorted_inds][1:9]

    # Realign eigenfunction signs for numerical comparison.
    for func, func_true in zip(edmd_eigenfunctions, eigenfunctions_true):
        if np.sign(func[0, 0]) != np.sign(func_true[0, 0]):
            func *= -1
        assert relative_error(func, func_true) < 1e-6


def test_operator():
    """
    Test that EDMD can sucessfully recover the true linear operator.
    Essentially tests that the eigenvalues and modes are correct.
    """
    edmd = EDMD(
        svd_rank=15,
        kernel_metric="poly",
        kernel_params={"gamma": 1, "coef0": 1, "degree": 4},
    )
    edmd.fit(X, Y)
    J_est = np.linalg.multi_dot(
        [edmd.modes, np.diag(edmd.eigs), np.linalg.pinv(edmd.modes)]
    )
    assert_allclose(J_est, J, atol=1e-13)


def test_reconstruction():
    """
    Test that EDMD can sucessfully reconstruct the forward-time system.
    """
    edmd = EDMD(
        svd_rank=15,
        kernel_metric="poly",
        kernel_params={"gamma": 1, "coef0": 1, "degree": 4},
    )
    edmd.fit(X, Y)
    assert_allclose(edmd.reconstructed_data, X2, atol=1e-13)


def test_kernel_errors():
    """
    Test that EDMD throws error upon initialization in the following scenarios:
        - kernel_metric is invalid
        - kernel_params isn't a dictionary
        - kernel_params contains invalid value types
        - kernel_params contains entries incompatible with the kernel
    """
    with raises(ValueError):
        EDMD(kernel_metric="rbf!")

    with raises(TypeError):
        EDMD(kernel_metric="rbf", kernel_params=3)

    with raises(ValueError):
        EDMD(kernel_metric="rbf", kernel_params={"gamma": "three"})

    with raises(ValueError):
        EDMD(kernel_metric="rbf", kernel_params={"gamma": 3, "degree": 5})


def test_svd_error():
    """
    Test that EDMD throws error upon initialization if svd_rank is zero.
    """
    with raises(ValueError):
        EDMD(svd_rank=0)


def test_atilde_error():
    """
    Test that EDMD throws error if a user asks for the operator matrix.
    """
    edmd = EDMD().fit(X, Y)

    with raises(ValueError):
        _ = edmd.operator.as_numpy_array

    with raises(ValueError):
        _ = edmd.operator.shape


def test_eigenfunction_error():
    """
    Test that EDMD throws an error if a user attempts to either
    - compute eigenfunctions prior to calling fit, or
    - compute eigenfunctions using input other than a 1-D numpy array.
    """
    edmd = EDMD()
    x_dummy = np.array([1, 1])

    with raises(ValueError):
        edmd.eigenfunctions(x_dummy)

    edmd.fit(X, Y)

    with raises(ValueError):
        edmd.eigenfunctions([1, 1])

    with raises(ValueError):
        edmd.eigenfunctions(x_dummy[None])



"""
Randomized pre-processing.
"""

from __future__ import annotations

import sys
from numbers import Number

import numpy as np

from pydmd.dmdbase import DMDBase
from pydmd.preprocessing.pre_post_processing import (
    PrePostProcessing,
    PrePostProcessingDMD,
)
from pydmd.utils import compute_rqb

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


def randomized_preprocessing(
    dmd: DMDBase,
    svd_rank: Number,
    oversampling: int = 10,
    power_iters: int = 2,
    test_matrix: np.ndarray = None,
    seed: int = None,
):
    """
    Randomized QB pre-processing.

    :param dmd: DMD instance to be wrapped.
    :param svd_rank: target rank of the input data.
    :param oversampling: amount to oversample beyond the target rank.
    :param power_iters: number of power iterations to perform.
    :param test_matrix: optional custom random test matrix.
    :param seed: optional random generator seed.
    """
    pre_post_processing = _RandomizedPrePostProcessing(
        svd_rank=svd_rank,
        oversampling=oversampling,
        power_iters=power_iters,
        test_matrix=test_matrix,
        seed=seed,
    )
    return PrePostProcessingDMD(dmd, pre_post_processing)


class _RandomizedPrePostProcessing(PrePostProcessing):
    def __init__(self, **kwargs):
        self._compute_rqb_kwargs = kwargs

    @override
    def pre_processing(self, X: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        Q = compute_rqb(X, **self._compute_rqb_kwargs).Q
        compression_matrix = Q.conj().T
        return compression_matrix, compression_matrix.dot(X)

    @override
    def post_processing(
        self, pre_processing_output: np.ndarray, Y: np.ndarray
    ) -> np.ndarray:
        return pre_processing_output.conj().T.dot(Y)



import logging

import numpy as np
from scipy.linalg import sqrtm

from .utils import compute_svd

logging.basicConfig(
    format="[%(filename)s:%(lineno)s - %(funcName)20s() ] %(message)s"
)


class DMDOperator:
    """
    Dynamic Mode Decomposition standard operator class. Non-standard ways of
    computing the low-rank Atilde operator should be coded into subclasses of
    this class.

    :param svd_rank: the rank for the truncation; If 0, the method computes the
        optimal rank and uses it for truncation; if positive interger, the
        method uses the argument for the truncation; if float between 0 and 1,
        the rank is the number of the biggest singular values that are needed
        to reach the 'energy' specified by `svd_rank`; if -1, the method does
        not compute truncation.
    :type svd_rank: int or float
    :param bool exact: flag to compute either exact DMD or projected DMD.
        Default is False.
    :param rescale_mode: Scale Atilde as shown in
        10.1016/j.jneumeth.2015.10.010 (section 2.4) before computing its
        eigendecomposition. None means no rescaling, 'auto' means automatic
        rescaling using singular values, otherwise the scaling factors.
    :type rescale_mode: {'auto'} or None or numpy.ndarray
    :param bool forward_backward: If True, the low-rank operator is computed
        like in fbDMD (reference: https://arxiv.org/abs/1507.02264). Default is
        False.
    :param sorted_eigs: Sort eigenvalues (and modes/dynamics accordingly) by
        magnitude if `sorted_eigs='abs'`, by real part (and then by imaginary
        part to break ties) if `sorted_eigs='real'`. Default: False.
    :type sorted_eigs: {'real', 'abs'} or False
    :param tikhonov_regularization: Tikhonov parameter for the regularization.
        If `None`, no regularization is applied, if `float`, it is used as the
        :math:`\lambda` tikhonov parameter.
    :type tikhonov_regularization: int or float
    """

    def __init__(
        self,
        svd_rank,
        exact,
        forward_backward,
        rescale_mode,
        sorted_eigs,
        tikhonov_regularization,
    ):
        self._exact = exact
        self._rescale_mode = rescale_mode
        self._svd_rank = svd_rank
        self._forward_backward = forward_backward
        self._sorted_eigs = sorted_eigs
        self._tikhonov_regularization = tikhonov_regularization
        self._norm_X = None

    def compute_operator(self, X, Y):
        """
        Compute the low-rank operator.

        :param numpy.ndarray X: matrix containing the snapshots x0,..x{n-1} by
            column.
        :param numpy.ndarray Y: matrix containing the snapshots x1,..x{n} by
            column.
        :return: the (truncated) left-singular vectors matrix, the (truncated)
            singular values array, the (truncated) right-singular vectors
            matrix of X.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray
        """

        U, s, V = compute_svd(X, self._svd_rank)

        if self._tikhonov_regularization is not None:
            self._norm_X = np.linalg.norm(X)
        atilde = self._least_square_operator(U, s, V, Y)

        if self._forward_backward:
            # b stands for "backward"
            bU, bs, bV = compute_svd(Y, svd_rank=len(s))
            atilde_back = self._least_square_operator(bU, bs, bV, X)
            atilde = sqrtm(atilde.dot(np.linalg.inv(atilde_back)))
            if hasattr(np, "complex256") and atilde.dtype == np.complex256:
                atilde = atilde.astype(np.complex128)
                msg = "Casting atilde from np.complex256 to np.complex128"
                logging.info(msg)

        if isinstance(self._rescale_mode, str) and self._rescale_mode == "auto":
            self._rescale_mode = s

        self._Atilde = atilde
        self._compute_eigenquantities()
        self._compute_modes(Y, U, s, V)

        return U, s, V

    @property
    def shape(self):
        """Shape of the operator"""
        return self.as_numpy_array.shape

    def __call__(self, snapshot_lowrank_modal_coefficients):
        """
        Apply the low-rank operator to a vector of the modal coefficients of a
        snapshot(s).

        :param numpy.ndarray snapshot_lowrank_modal_coefficients: low-rank
            representation (in modal coefficients) of a snapshot x{n}.
        :return: low-rank representation (in modal coefficients) of x{n+1}.
        :rtype: numpy.ndarray
        """

        return self._Atilde.dot(snapshot_lowrank_modal_coefficients)

    @property
    def eigenvalues(self):
        if not hasattr(self, "_eigenvalues"):
            raise ValueError("You need to call fit before")
        return self._eigenvalues

    @property
    def eigenvectors(self):
        if not hasattr(self, "_eigenvectors"):
            raise ValueError("You need to call fit before")
        return self._eigenvectors

    @property
    def modes(self):
        if not hasattr(self, "_modes"):
            raise ValueError("You need to call fit before")
        return self._modes

    @property
    def Lambda(self):
        if not hasattr(self, "_Lambda"):
            raise ValueError("You need to call fit before")
        return self._Lambda

    @property
    def as_numpy_array(self):
        if not hasattr(self, "_Atilde") or self._Atilde is None:
            raise ValueError("You need to call fit before")
        else:
            return self._Atilde

    def _least_square_operator(self, U, s, V, Y):
        """
        Private method that computes the lowrank operator from the singular
        value decomposition of matrix X and the matrix Y.

        .. math::

            \\mathbf{\\tilde{A}} =
            \\mathbf{U}^* \\mathbf{Y} \\mathbf{X}^\\dagger \\mathbf{U} =
            \\mathbf{U}^* \\mathbf{Y} \\mathbf{V} \\mathbf{S}^{-1}

        :param numpy.ndarray U: 2D matrix that contains the left-singular
            vectors of X, stored by column.
        :param numpy.ndarray s: 1D array that contains the singular values of
            X.
        :param numpy.ndarray V: 2D matrix that contains the right-singular
            vectors of X, stored by row.
        :param numpy.ndarray Y: input matrix Y.
        :return: the lowrank operator
        :rtype: numpy.ndarray
        """
        if self._tikhonov_regularization is not None:
            s = (
                s**2 + self._tikhonov_regularization * self._norm_X
            ) * np.reciprocal(s)
        return np.linalg.multi_dot([U.T.conj(), Y, V]) * np.reciprocal(s)

    def _compute_eigenquantities(self):
        """
        Private method that computes eigenvalues and eigenvectors of the
        low-dimensional operator, scaled according to self._rescale_mode.
        """

        if self._rescale_mode is None:
            # scaling isn't required
            Ahat = self._Atilde
        elif isinstance(self._rescale_mode, np.ndarray):
            if len(self._rescale_mode) != self.as_numpy_array.shape[0]:
                raise ValueError(
                    """Scaling by an invalid number of
                        coefficients"""
                )
            scaling_factors_array = self._rescale_mode

            factors_inv_sqrt = np.diag(np.power(scaling_factors_array, -0.5))
            factors_sqrt = np.diag(np.power(scaling_factors_array, 0.5))

            # if an index is 0, we get inf when taking the reciprocal
            for idx, item in enumerate(scaling_factors_array):
                if item == 0:
                    factors_inv_sqrt[idx] = 0

            Ahat = np.linalg.multi_dot(
                [factors_inv_sqrt, self.as_numpy_array, factors_sqrt]
            )
        else:
            raise ValueError(
                "Invalid value for rescale_mode: {} of type {}".format(
                    self._rescale_mode, type(self._rescale_mode)
                )
            )

        self._eigenvalues, self._eigenvectors = np.linalg.eig(Ahat)

        if self._sorted_eigs is not False and self._sorted_eigs is not None:
            if self._sorted_eigs == "abs":

                def k(tp):
                    return abs(tp[0])

            elif self._sorted_eigs == "real":

                def k(tp):
                    eig = tp[0]
                    if isinstance(eig, complex):
                        return (eig.real, eig.imag)
                    return (eig.real, 0)

            else:
                raise ValueError(
                    "Invalid value for sorted_eigs: {}".format(
                        self._sorted_eigs
                    )
                )

            # each column is an eigenvector, therefore we take the
            # transpose to associate each row (former column) to an
            # eigenvalue before sorting
            a, b = zip(
                *sorted(zip(self._eigenvalues, self._eigenvectors.T), key=k)
            )
            self._eigenvalues = np.array([eig for eig in a])
            # we restore the original condition (eigenvectors in columns)
            self._eigenvectors = np.array([vec for vec in b]).T

    def _compute_modes(self, Y, U, Sigma, V):
        """
        Private method that computes eigenvalues and eigenvectors of the
        high-dimensional operator (stored in self.modes and self.Lambda).

        :param numpy.ndarray Y: matrix containing the snapshots x1,..x{n} by
            column.
        :param numpy.ndarray U: (truncated) left singular vectors of X
        :param numpy.ndarray Sigma: (truncated) singular values of X
        :param numpy.ndarray V: (truncated) right singular vectors of X
        """

        if self._rescale_mode is None:
            W = self.eigenvectors
        else:
            # compute W as shown in arXiv:1409.5496 (section 2.4)
            factors_sqrt = np.diag(np.power(self._rescale_mode, 0.5))
            W = factors_sqrt.dot(self.eigenvectors)

        # compute the eigenvectors of the high-dimensional operator
        if self._exact:
            if self._tikhonov_regularization is not None:
                Sigma = (
                    Sigma**2 + self._tikhonov_regularization * self._norm_X
                ) * np.reciprocal(Sigma)
            high_dimensional_eigenvectors = (
                Y.dot(V) * np.reciprocal(Sigma)
            ).dot(W)
        else:
            high_dimensional_eigenvectors = U.dot(W)

        # eigenvalues are the same of lowrank
        high_dimensional_eigenvalues = self.eigenvalues

        self._modes = high_dimensional_eigenvectors
        self._Lambda = high_dimensional_eigenvalues



import numpy as np
from pytest import raises

from pydmd import DMD
from pydmd.dmdbase import DMDBase
from pydmd.snapshots import Snapshots

# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")


def test_svd_rank_default():
    dmd = DMDBase()
    assert dmd.operator._svd_rank == 0


def test_svd_rank():
    dmd = DMDBase(svd_rank=3)
    assert dmd.operator._svd_rank == 3


def test_tlsq_rank_default():
    dmd = DMDBase()
    assert dmd._tlsq_rank == 0


def test_tlsq_rank():
    dmd = DMDBase(tlsq_rank=2)
    assert dmd._tlsq_rank == 2


def test_exact_default():
    dmd = DMDBase()
    assert dmd.operator._exact == False


def test_exact():
    dmd = DMDBase(exact=True)
    assert dmd.operator._exact == True


def test_opt_default():
    dmd = DMDBase()
    assert dmd._opt == False


def test_opt():
    dmd = DMDBase(opt=True)
    assert dmd._opt == True


def test_fit():
    dmd = DMDBase(exact=False)
    with raises(NotImplementedError):
        dmd.fit(sample_data)


def test_advanced_snapshot_parameter2():
    dmd = DMDBase(opt=5)
    assert dmd._opt == 5


def test_translate_tpow_positive():
    dmd = DMDBase(opt=4)

    assert dmd._translate_eigs_exponent(10) == 6
    assert dmd._translate_eigs_exponent(0) == -4


def test_translate_tpow_negative():
    dmd = DMDBase(opt=-1)
    dmd._snapshots_holder = Snapshots(sample_data)

    assert dmd._translate_eigs_exponent(10) == 10 - (sample_data.shape[1] - 1)
    assert dmd._translate_eigs_exponent(0) == 1 - sample_data.shape[1]


def test_translate_tpow_vector():
    dmd = DMDBase(opt=-1)
    dmd._snapshots_holder = Snapshots(sample_data)

    tpow = np.ndarray([0, 1, 2, 3, 5, 6, 7, 11])
    for idx, x in enumerate(dmd._translate_eigs_exponent(tpow)):
        assert x == dmd._translate_eigs_exponent(tpow[idx])


def test_sorted_eigs_default():
    dmd = DMDBase()
    assert dmd.operator._sorted_eigs == False


def test_sorted_eigs_param():
    dmd = DMDBase(sorted_eigs="real")
    assert dmd.operator._sorted_eigs == "real"


def test_dmd_time_wrong_key():
    dmd = DMD(svd_rank=10)
    dmd.fit(sample_data)

    with raises(KeyError):
        dmd.dmd_time["tstart"] = 10



import pkgutil
from os import path, walk


def test_import_dm_1():
    import pydmd as dm

    dmd = dm.dmd.DMDBase()


def test_import_dm_2():
    import pydmd as dm

    dmd = dm.dmd.DMD()


def test_import_dm_3():
    import pydmd as dm

    dmd = dm.fbdmd.FbDMD()



import numpy as np
from pytest import raises
from scipy.integrate import solve_ivp

from pydmd.bopdmd import BOPDMD
from pydmd.dmd import DMD


def simulate_z(t_eval):
    """
    Given a time vector t_eval = t1, t2, ..., evaluates and returns
    the snapshots z(t1), z(t2), ... as columns of the matrix Z.
    Simulates data z given by the system of ODEs
        z' = Az
    where A = [1 -2; 1 -1] and z_0 = [1, 0.1].
    """

    def ode_sys(zt, z):
        z1, z2 = z
        return [z1 - 2 * z2, z1 - z2]

    # Set integrator keywords to replicate odeint defaults.
    integrator_keywords = {}
    integrator_keywords["rtol"] = 1e-12
    integrator_keywords["method"] = "LSODA"
    integrator_keywords["atol"] = 1e-12

    sol = solve_ivp(
        ode_sys,
        [t_eval[0], t_eval[-1]],
        [1.0, 0.1],
        t_eval=t_eval,
        **integrator_keywords,
    )

    return sol.y


def differentiate(X, dt):
    """
    Method for performing 2nd order centered finite difference. Assumes the
    input matrix X is 2-D, with uniformly-sampled snapshots filling each
    column. Requires dt, which is the time step between each snapshot.
    """
    if not isinstance(X, np.ndarray) or X.ndim != 2:
        raise ValueError("Please ensure that input data is a 2D array.")

    X_prime = np.empty(X.shape)
    X_prime[:, 1:-1] = (X[:, 2:] - X[:, :-2]) / (2 * dt)
    X_prime[:, 0] = (-3 * X[:, 0] + 4 * X[:, 1] - X[:, 2]) / (2 * dt)
    X_prime[:, -1] = (3 * X[:, -1] - 4 * X[:, -2] + X[:, -3]) / (2 * dt)

    return X_prime


# Simulate ODE system data.
t = np.arange(2000) * 0.01
Z = simulate_z(t)
Z_dot = differentiate(Z, t[1] - t[0])

# Additionally re-obtain DMD testing data.
# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")
sample_data_1 = sample_data[:, :-1]
sample_data_2 = sample_data[:, 1:]


def assert_equal_models(dmd, dmd_true, rtol=1e-07, atol=0):
    """
    Helper method for comparing two DMD models. Ensures that the eigs, modes,
    amplitudes, and Atilde operators agree up to the given tolerance values.
    """
    # Compare eigenvalues.
    np.testing.assert_allclose(dmd.eigs, dmd_true.eigs, rtol=rtol, atol=atol)
    # Compare modes.
    np.testing.assert_allclose(dmd.modes, dmd_true.modes, rtol=rtol, atol=atol)
    # Compare amplitudes.
    np.testing.assert_allclose(
        dmd.amplitudes, dmd_true.amplitudes, rtol=rtol, atol=atol
    )
    # Compare Atilde matrices.
    np.testing.assert_allclose(
        dmd.operator.as_numpy_array,
        dmd_true.operator.as_numpy_array,
        rtol=rtol,
        atol=atol,
    )


def test_shape_error_1():
    """
    Checks that an error is thrown when X and Y
    don't contain the same number of snapshots.
    """
    dmd = DMD()
    with raises(ValueError):
        dmd.fit(X=sample_data, Y=sample_data[:, 1:])


def test_shape_error_2():
    """
    Checks that an error is thrown when the
    snapshots in X and Y aren't the same size.
    """
    dmd = DMD()
    with raises(ValueError):
        dmd.fit(X=sample_data, Y=sample_data[1:])


def test_time_dicts():
    """
    Checks that the default time dictionaries contain 0, 1, ..., m-1
    when data matrices X, Y containing m snapshots each are given.
    """
    dmd = DMD()
    dmd.fit(X=sample_data_1, Y=sample_data_2)
    expected_dict = {"dt": 1, "t0": 0, "tend": 13}
    np.testing.assert_equal(dmd.original_time, expected_dict)
    np.testing.assert_equal(dmd.dmd_time, expected_dict)


def test_equal_models_default():
    """
    Checks that a DMD model given X=data, Y=None is qualitatively the same as
    a DMD model given X=data[:, :-1], Y=data[:, 1:] using default parameters.
    """
    dmd = DMD(svd_rank=2)
    dmd.fit(X=sample_data)

    dmd_xy = DMD(svd_rank=2)
    dmd_xy.fit(X=sample_data_1, Y=sample_data_2)

    assert_equal_models(dmd_xy, dmd)


def test_equal_models_exact():
    """
    Checks that a DMD model given X=data, Y=None is qualitatively the same as
    a DMD model given X=data[:, :-1], Y=data[:, 1:] using exact=True.
    """
    dmd = DMD(svd_rank=2, exact=True)
    dmd.fit(X=sample_data)

    dmd_xy = DMD(svd_rank=2, exact=True)
    dmd_xy.fit(X=sample_data_1, Y=sample_data_2)

    assert_equal_models(dmd_xy, dmd)


def test_equal_models_opt():
    """
    Checks that a DMD model given X=data, Y=None is almost qualitatively the
    same as a DMD model given X=data[:, :-1], Y=data[:, 1:] using opt=True.
    """
    dmd = DMD(svd_rank=2, opt=True)
    dmd.fit(X=sample_data)

    dmd_xy = DMD(svd_rank=2, opt=True)
    dmd_xy.fit(X=sample_data_1, Y=sample_data_2)

    assert_equal_models(dmd_xy, dmd, rtol=0.05)


def test_equal_models_opt_exact():
    """
    Checks that a DMD model given X=data, Y=None is qualitatively the same as
    a model given X=data[:, :-1], Y=data[:, 1:] using opt=True, exact=True.
    """
    dmd = DMD(svd_rank=2, opt=True, exact=True)
    dmd.fit(X=sample_data)

    dmd_xy = DMD(svd_rank=2, opt=True, exact=True)
    dmd_xy.fit(X=sample_data_1, Y=sample_data_2)

    assert_equal_models(dmd_xy, dmd)


def test_time_shifted_model():
    """
    Checks that a DMD model fitted with X=data, Y=shifted(data) and missing
    data is able to recover the diagnostics obtained using all of the data.
    """
    uneven_indices = np.delete(np.arange(len(t)), np.arange(1000)[1::2])
    uneven_indices = uneven_indices[:-1]

    Z_uneven_1 = Z[:, uneven_indices]
    Z_uneven_2 = Z[:, uneven_indices + 1]

    dmd = DMD(svd_rank=2)
    dmd.fit(X=Z)

    dmd_xy = DMD(svd_rank=2)
    dmd_xy.fit(X=Z_uneven_1, Y=Z_uneven_2)

    assert_equal_models(dmd_xy, dmd, rtol=0.01)


def test_time_derivative_model():
    """
    Checks that a DMD model given X=data, Y=derivative(data) yields
    the same operator as a BOPDMD model given the same data set.
    """
    dmd = DMD(svd_rank=2)
    dmd.fit(X=Z, Y=Z_dot)

    bopdmd = BOPDMD(svd_rank=2)
    bopdmd.fit(Z, t)

    np.testing.assert_allclose(
        dmd.operator.as_numpy_array,
        bopdmd.operator.as_numpy_array,
        rtol=0.01,
    )



from builtins import range

import numpy as np
from pytest import raises

from pydmd import HankelDMD

# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")


def create_noisy_data():
    mu = 0.0
    sigma = 0.0  # noise standard deviation
    m = 100  # number of snapshot
    noise = np.random.normal(mu, sigma, m)  # gaussian noise
    A = np.array([[1.0, 1.0], [-1.0, 2.0]])
    A /= np.sqrt(3)
    n = 2
    X = np.zeros((n, m))
    X[:, 0] = np.array([0.5, 1.0])
    # evolve the system and perturb the data with noise
    for k in range(1, m):
        X[:, k] = A.dot(X[:, k - 1])
        X[:, k - 1] += noise[k - 1]
    return X


noisy_data = create_noisy_data()


def test_shape():
    dmd = HankelDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert dmd.modes.shape[1] == sample_data.shape[1] - 1


def test_truncation_shape():
    dmd = HankelDMD(svd_rank=3)
    dmd.fit(X=sample_data)
    assert dmd.modes.shape[1] == 3


def test_rank():
    dmd = HankelDMD(svd_rank=0.9)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 2


def test_Atilde_shape():
    dmd = HankelDMD(svd_rank=3)
    dmd.fit(X=sample_data)
    assert dmd.operator.as_numpy_array.shape == (
        dmd.operator._svd_rank,
        dmd.operator._svd_rank,
    )


def test_d():
    single_data = np.sin(np.linspace(0, 10, 100))[None]
    dmd = HankelDMD(svd_rank=0, d=50, opt=True)
    dmd.fit(single_data)
    np.testing.assert_array_almost_equal(
        dmd.reconstructed_data.real, single_data, decimal=1
    )  # TODO poor accuracy using projected modes
    dmd = HankelDMD(svd_rank=-1, d=50, opt=True, exact=True)
    dmd.fit(single_data)
    assert np.allclose(dmd.reconstructed_data.flatten().real, single_data)


def test_Atilde_values():
    dmd = HankelDMD(svd_rank=2)
    dmd.fit(X=sample_data)
    exact_atilde = np.array(
        [
            [-0.70558526 + 0.67815084j, 0.22914898 + 0.20020143j],
            [0.10459069 + 0.09137814j, -0.57730040 + 0.79022994j],
        ]
    )
    np.testing.assert_allclose(exact_atilde, dmd.operator.as_numpy_array)


def test_eigs_1():
    dmd = HankelDMD(svd_rank=-1)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 14


def test_eigs_2():
    dmd = HankelDMD(svd_rank=5)
    dmd.fit(X=sample_data)
    assert len(dmd.eigs) == 5


def test_eigs_3():
    dmd = HankelDMD(svd_rank=2)
    dmd.fit(X=sample_data)
    expected_eigs = np.array(
        [
            -8.09016994e-01 + 5.87785252e-01j,
            -4.73868662e-01 + 8.80595532e-01j,
        ]
    )
    np.testing.assert_almost_equal(dmd.eigs, expected_eigs, decimal=6)


def test_dynamics_1():
    dmd = HankelDMD(svd_rank=5)
    dmd.fit(X=sample_data)
    assert dmd.dynamics.shape == (5, sample_data.shape[1])


def test_dynamics_2():
    dmd = HankelDMD(svd_rank=1)
    dmd.fit(X=sample_data)
    expected_dynamics = np.array(
        [
            [
                -2.20639502 - 9.10168802e-16j,
                1.55679980 - 1.49626864e00j,
                -0.08375915 + 2.11149018e00j,
                -1.37280962 - 1.54663768e00j,
                2.01748787 + 1.60312745e-01j,
                -1.53222592 + 1.25504678e00j,
                0.23000498 - 1.92462280e00j,
                1.14289644 + 1.51396355e00j,
                -1.83310653 - 2.93174173e-01j,
                1.49222925 - 1.03626336e00j,
                -0.35015209 + 1.74312867e00j,
                -0.93504202 - 1.46738182e00j,
                1.65485808 + 4.01263449e-01j,
                -1.43976061 + 8.39117825e-01j,
                0.44682540 - 1.56844403e00j,
            ]
        ]
    )
    np.testing.assert_allclose(dmd.dynamics, expected_dynamics)


def test_dynamics_opt_1():
    dmd = HankelDMD(svd_rank=5, opt=True)
    dmd.fit(X=sample_data)
    assert dmd.dynamics.shape == (5, sample_data.shape[1])


def test_dynamics_opt_2():
    dmd = HankelDMD(svd_rank=1, opt=True, exact=False)
    dmd.fit(X=sample_data)
    expected_dynamics = np.array(
        [
            [
                -4.609718826226513 - 6.344781724790875j,
                7.5552686987577165 + 1.3506997434096375j,
                -6.246864367654589 + 4.170577993207872j,
                1.5794144248628537 - 7.179014663490048j,
                3.754043295828462 + 6.13648812118528j,
                -6.810262177959786 - 1.7840079278093528j,
                6.015047060133346 - 3.35961532862783j,
                -1.9658025630719695 + 6.449604262000736j,
                -2.9867632454837936 - 5.8838563367460734j,
                6.097558230017521 + 2.126086276430128j,
                -5.7441543819530265 + 2.6349291080417103j,
                2.266111252852836 - 5.7545702519088895j,
                2.303531963541068 + 5.597105176945707j,
                -5.421019770795679 - 2.3870927539102658j,
                5.443800581850978 - 1.9919716610066682j,
            ]
        ]
    )
    np.testing.assert_allclose(dmd.dynamics, expected_dynamics)


def test_reconstructed_data():
    dmd = HankelDMD()
    dmd.fit(X=sample_data)
    dmd_data = dmd.reconstructed_data
    np.testing.assert_allclose(dmd_data, sample_data)


def test_original_time():
    dmd = HankelDMD(svd_rank=2)
    dmd.fit(X=sample_data)
    expected_dict = {"dt": 1, "t0": 0, "tend": 14}
    np.testing.assert_equal(dmd.original_time, expected_dict)


def test_original_timesteps():
    dmd = HankelDMD()
    dmd.fit(X=sample_data)
    np.testing.assert_allclose(
        dmd.original_timesteps, np.arange(sample_data.shape[1])
    )


def test_dmd_time_1():
    dmd = HankelDMD(svd_rank=2)
    dmd.fit(X=sample_data)
    expected_dict = {"dt": 1, "t0": 0, "tend": 14}
    np.testing.assert_equal(dmd.dmd_time, expected_dict)


def test_dmd_time_2():
    dmd = HankelDMD()
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 10
    dmd.dmd_time["tend"] = 14
    expected_data = sample_data[:, -5:]
    np.testing.assert_allclose(dmd.reconstructed_data, expected_data)


def test_dmd_time_3():
    dmd = HankelDMD()
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 8
    dmd.dmd_time["tend"] = 11
    expected_data = sample_data[:, 8:12]
    np.testing.assert_allclose(dmd.reconstructed_data, expected_data)


def test_dmd_time_4():
    dmd = HankelDMD(svd_rank=3)
    dmd.fit(X=sample_data)
    dmd.dmd_time["t0"] = 20
    dmd.dmd_time["tend"] = 20
    expected_data = np.array(
        [
            [-7.29383297e00 - 4.90248179e-14j],
            [-5.69109796e00 - 2.74068833e00j],
            [3.38410649e-83 + 3.75677740e-83j],
        ]
    )
    np.testing.assert_almost_equal(dmd.dynamics, expected_data, decimal=6)


def test_dmd_time_5():
    x = np.linspace(0, 10, 64)[None]
    y = np.cos(x) * np.sin(np.cos(x)) + np.cos(x * 0.2)

    dmd = HankelDMD(svd_rank=-1, exact=True, opt=True, d=30)
    dmd.fit(y)

    dmd.original_time["dt"] = dmd.dmd_time["dt"] = x[0, 1] - x[0, 0]
    dmd.original_time["t0"] = dmd.dmd_time["t0"] = x[0, 0]
    dmd.original_time["tend"] = dmd.dmd_time["tend"] = x[0, -1]

    assert dmd.reconstructed_data.shape == (1, 64)


def test_sorted_eigs_default():
    dmd = HankelDMD()
    assert dmd.operator._sorted_eigs == False


def test_sorted_eigs_param():
    dmd = HankelDMD(sorted_eigs="real")
    assert dmd.operator._sorted_eigs == "real"


def test_reconstruction_method_constructor_error():
    with raises(ValueError):
        HankelDMD(reconstruction_method=[1, 2, 3], d=4)

    with raises(ValueError):
        HankelDMD(reconstruction_method=np.array([1, 2, 3]), d=4)

    with raises(ValueError):
        HankelDMD(reconstruction_method=np.array([[1, 2, 3], [3, 4, 5]]), d=3)


def test_reconstruction_method_default_constructor():
    assert HankelDMD()._reconstruction_method == "first"


def test_nonan_nomask():
    dmd = HankelDMD(d=3)
    dmd.fit(X=sample_data)
    rec = dmd.reconstructed_data

    assert not isinstance(rec, np.ma.MaskedArray)
    assert not np.nan in rec


def test_hankeldmd_timesteps():
    x = np.linspace(0, 10, 64)[None]
    arr = np.cos(x) * np.sin(np.cos(x)) + np.cos(x * 0.2)
    dmd = HankelDMD(svd_rank=1, exact=True, opt=True, d=30).fit(arr)
    assert len(dmd.dmd_timesteps) == 64


def test_first_occurences():
    x = np.linspace(0, 10, 64)[None]
    arr = np.cos(x) * np.sin(np.cos(x)) + np.cos(x * 0.2)
    dmd = HankelDMD(svd_rank=1, exact=True, opt=True, d=3).fit(arr)
    assert dmd._hankel_first_occurrence(0) == 0
    assert dmd._hankel_first_occurrence(1) == 0
    assert dmd._hankel_first_occurrence(2) == 0
    assert dmd._hankel_first_occurrence(3) == 1
    assert dmd._hankel_first_occurrence(4) == 2
    assert dmd._hankel_first_occurrence(5) == 3

    dmd.dmd_time["tend"] = 100
    assert dmd._hankel_first_occurrence(100) == 98

    # change scale
    dmd.dmd_time["t0"] = dmd.original_time["t0"] = x[0, 0]
    dmd.dmd_time["tend"] = dmd.original_time["tend"] = x[0, -1]
    dmd.dmd_time["dt"] = dmd.original_time["dt"] = x[0, 1] - x[0, 0]

    assert dmd._hankel_first_occurrence(x[0, 0]) == 0
    assert dmd._hankel_first_occurrence(x[0, 1]) == 0
    assert dmd._hankel_first_occurrence(x[0, 2]) == 0
    assert dmd._hankel_first_occurrence(x[0, 3]) == 1
    assert dmd._hankel_first_occurrence(x[0, -1] + dmd.dmd_time["dt"]) == 62

    dmd.dmd_time["t0"] = x[0, x.shape[-1] // 2]
    dmd.dmd_time["tend"] = x[0, -1] + dmd.dmd_time["dt"] * 20

    assert (
        dmd._hankel_first_occurrence(dmd.dmd_time["t0"]) == x.shape[-1] // 2 - 2
    )


def test_update_sub_dmd_time():
    dmd = HankelDMD()
    x = np.linspace(0, 10, 64)[None]
    arr = np.cos(x) * np.sin(np.cos(x)) + np.cos(x * 0.2)
    dmd = HankelDMD(svd_rank=1, exact=True, opt=True, d=3).fit(arr)

    dmd.dmd_time["tend"] += dmd.dmd_time["dt"] * 20
    dmd._update_sub_dmd_time()

    # assert that the dt for the sub_dmd is always 1
    assert dmd._sub_dmd.dmd_time["dt"] == 1
    assert dmd._sub_dmd.original_time["dt"] == 1

    assert (
        dmd._sub_dmd.dmd_time["tend"] == dmd._sub_dmd.original_time["tend"] + 20
    )


def test_hankel_2d():
    def fnc(x):
        return np.cos(x) * np.sin(np.cos(x)) + np.cos(x * 0.2)

    x = np.linspace(0, 10, 64)
    snapshots = np.vstack([fnc(x), -fnc(x)])

    dmd = HankelDMD(svd_rank=0, exact=True, opt=True, d=30).fit(snapshots)

    dmd.original_time["dt"] = dmd.dmd_time["dt"] = x[1] - x[0]
    dmd.original_time["t0"] = dmd.dmd_time["t0"] = x[0]
    dmd.original_time["tend"] = dmd.dmd_time["tend"] = x[-1]

    dmd.dmd_time["t0"] = x[len(x) // 2]
    dmd.dmd_time["tend"] = x[-1] + dmd.dmd_time["dt"] * 20

    assert len(dmd.dmd_timesteps) == dmd.reconstructed_data.shape[1]

    np.testing.assert_allclose(
        dmd.reconstructed_data,
        np.vstack([fnc(dmd.dmd_timesteps), -fnc(dmd.dmd_timesteps)]),
    )


def test_get_bitmask_default():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)
    assert np.all(dmd.modes_activation_bitmask == True)


def test_set_bitmask():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)

    new_bitmask = np.full(len(dmd.amplitudes), True, dtype=bool)
    new_bitmask[[0]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes_activation_bitmask[0] == False
    assert np.all(dmd.modes_activation_bitmask[1:] == True)


def test_not_fitted_get_bitmask_raises():
    dmd = HankelDMD(svd_rank=-1, d=5)
    with raises(RuntimeError):
        print(dmd.modes_activation_bitmask)


def test_not_fitted_set_bitmask_raises():
    dmd = HankelDMD(svd_rank=-1, d=5)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, True, dtype=bool)


def test_raise_wrong_dtype_bitmask():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)
    with raises(RuntimeError):
        dmd.modes_activation_bitmask = np.full(3, 0.1)


def test_fitted():
    dmd = HankelDMD(svd_rank=-1, d=5)
    assert not dmd.fitted
    dmd.fit(X=sample_data)
    assert dmd.fitted


def test_bitmask_amplitudes():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)

    old_n_amplitudes = dmd.amplitudes.shape[0]
    retained_amplitudes = np.delete(dmd.amplitudes, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.amplitudes.shape[0] == old_n_amplitudes - 2
    np.testing.assert_almost_equal(dmd.amplitudes, retained_amplitudes)


def test_bitmask_eigs():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)

    old_n_eigs = dmd.eigs.shape[0]
    retained_eigs = np.delete(dmd.eigs, [0, -1])

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.eigs.shape[0] == old_n_eigs - 2
    np.testing.assert_almost_equal(dmd.eigs, retained_eigs)


def test_bitmask_modes():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)

    old_n_modes = dmd.modes.shape[1]
    retained_modes = np.delete(dmd.modes, [0, -1], axis=1)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    assert dmd.modes.shape[1] == old_n_modes - 2
    np.testing.assert_almost_equal(dmd.modes, retained_modes)


def test_reconstructed_data_with_bitmask():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)

    new_bitmask = np.full(dmd.amplitudes.shape[0], True, dtype=bool)
    new_bitmask[[0, -1]] = False
    dmd.modes_activation_bitmask = new_bitmask

    dmd.reconstructed_data
    assert True


def test_correct_amplitudes():
    dmd = HankelDMD(svd_rank=-1, d=5)
    dmd.fit(X=sample_data)
    np.testing.assert_array_almost_equal(dmd.amplitudes, dmd._sub_dmd._b)


def test_raises_not_enough_snapshots():
    dmd = HankelDMD(svd_rank=-1, d=5)
    with raises(
        ValueError,
        match="The number of snapshots provided is not enough for d=5.\nExpected at least d.",
    ):
        dmd.fit(np.ones((20, 4)))
    with raises(ValueError, match="Received only one time snapshot."):
        dmd.fit(np.ones((20, 5)))
    dmd.fit(np.ones((20, 6)))


def test_hankeldmd_1d():
    dmd = HankelDMD(svd_rank=-1, d=5)
    data = np.pi * np.cos(np.linspace(0, 2, 100))[None]
    dmd.fit(data)
    np.testing.assert_allclose(data, dmd.reconstructed_data)



"""
PyDMD init
"""

__all__ = [
    "dmdbase",
    "dmd",
    "fbdmd",
    "mrdmd",
    "cdmd",
    "hodmd",
    "dmdc",
    "optdmd",
    "hankeldmd",
    "rdmd",
    "havok",
    "bopdmd",
    "pidmd",
    "edmd",
    "varprodmd",
    "lando",
]


from .bopdmd import BOPDMD
from .cdmd import CDMD
from .dmd import DMD
from .dmd_modes_tuner import ModesTuner
from .dmdbase import DMDBase
from .dmdc import DMDc
from .edmd import EDMD
from .fbdmd import FbDMD
from .hankeldmd import HankelDMD
from .havok import HAVOK
from .hodmd import HODMD
from .lando import LANDO
from .mrdmd import MrDMD
from .optdmd import OptDMD
from .paramdmd import ParametricDMD
from .pidmd import PiDMD
from .preprocessing import PrePostProcessingDMD
from .rdmd import RDMD
from .spdmd import SpDMD
from .subspacedmd import SubspaceDMD
from .varprodmd import VarProDMD



from copy import deepcopy

import numpy as np
import pytest
from ezyrb import POD, RBF
from pytest import param, raises

from pydmd import (
    CDMD,
    DMD,
    HODMD,
    DMDBase,
    DMDc,
    FbDMD,
    HankelDMD,
    MrDMD,
    OptDMD,
    ParametricDMD,
    SpDMD,
)
from pydmd.dmd_modes_tuner import (
    ModesSelectors,
    ModesTuner,
    select_modes,
    selectors,
    stabilize_modes,
)

# 15 snapshot with 400 data. The matrix is 400x15 and it contains
# the following data: f1 + f2 where
# f1 = lambda x,t: sech(x+3)*(1.*np.exp(1j*2.3*t))
# f2 = lambda x,t: (sech(x)*np.tanh(x))*(2.*np.exp(1j*2.8*t))
sample_data = np.load("tests/test_datasets/input_sample.npy")


class FakeDMDOperator:
    def __init__(self):
        self.as_numpy_array = np.ones(10)


def test_select_modes():
    def stable_modes(dmd_object):
        toll = 1e-3
        return np.abs(np.abs(dmd_object.eigs) - 1) < toll

    dmd = DMD(svd_rank=10)
    dmd.fit(sample_data)
    dmdc = deepcopy(dmd)

    exp = dmd.reconstructed_data
    select_modes(dmd, stable_modes)
    np.testing.assert_array_almost_equal(exp, dmd.reconstructed_data)

    assert len(dmd.eigs) < len(dmdc.eigs)
    assert dmd.modes.shape[1] < dmdc.modes.shape[1]
    assert len(dmd.amplitudes) < len(dmdc.amplitudes)


def test_select_modes_nullified_indexes():
    def stable_modes(dmd_object):
        toll = 1e-3
        return np.abs(np.abs(dmd_object.eigs) - 1) < toll

    dmd = DMD(svd_rank=10)
    dmd.fit(sample_data)
    dmdc = deepcopy(dmd)

    _, cut_indexes = select_modes(
        dmd, stable_modes, nullify_amplitudes=False, return_indexes=True
    )
    noncut_indexes = list(set(range(len(dmdc.eigs))) - set(cut_indexes))

    assert dmd.amplitudes.shape == dmdc[noncut_indexes].amplitudes.shape


def test_select_modes_index():
    fake_dmd_operator = FakeDMDOperator()
    fake_dmd = DMD()

    eigs = np.array([1 + 1e-4, 2, 1 - 1e-2, 5, 1, 1 - 0.5 * 1e-3])

    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "_Lambda", np.zeros(len(eigs)))
    # these are DMD eigenvectors, but we do not care in this test
    setattr(fake_dmd_operator, "_eigenvectors", np.zeros((1, len(eigs))))
    setattr(fake_dmd_operator, "_modes", np.zeros((1, len(eigs))))
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))

    setattr(fake_dmd, "_Atilde", fake_dmd_operator)
    setattr(fake_dmd, "_b", np.zeros(len(eigs)))

    _, idx = select_modes(
        fake_dmd,
        ModesSelectors.stable_modes(max_distance_from_unity=1e-3),
        return_indexes=True,
    )
    np.testing.assert_array_equal(idx, [1, 2, 3])

    assert fake_dmd.modes.shape[1] == 3
    assert len(fake_dmd.eigs) == 3
    assert len(fake_dmd.amplitudes) == 3


def test_select_modes_index_and_deepcopy():
    fake_dmd_operator = FakeDMDOperator()
    fake_dmd = DMD()

    eigs = np.array([1 + 1e-4, 2, 1 - 1e-2, 5, 1, 1 - 0.5 * 1e-3])

    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(fake_dmd_operator, "_Lambda", np.zeros(len(eigs)))
    # these are DMD eigenvectors, but we do not care in this test
    setattr(fake_dmd_operator, "_eigenvectors", np.zeros((1, len(eigs))))
    setattr(fake_dmd_operator, "_modes", np.zeros((1, len(eigs))))
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
    setattr(fake_dmd, "_b", np.zeros(len(eigs)))

    setattr(fake_dmd, "_Atilde", fake_dmd_operator)

    dmd2, idx = select_modes(
        fake_dmd,
        ModesSelectors.stable_modes(max_distance_from_unity=1e-3),
        in_place=False,
        return_indexes=True,
    )
    np.testing.assert_array_equal(idx, [1, 2, 3])

    assert len(fake_dmd.eigs) == 6
    assert fake_dmd.modes.shape[1] == 6
    assert len(fake_dmd.amplitudes) == 6

    assert len(dmd2.eigs) == 3
    assert dmd2.modes.shape[1] == 3
    assert len(dmd2.amplitudes) == 3


def test_stable_modes_both():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array([1 + 1e-4, 2, 1 - 1e-2, 5, 1, 1 - 0.5 * 1e-3]),
    )

    expected_result = np.array([False for _ in range(6)])
    expected_result[[0, 4, 5]] = True

    assert all(
        ModesSelectors.stable_modes(max_distance_from_unity=1e-3)(fake_dmd)
        == expected_result
    )


def test_stable_modes_outside_only():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array([1 + 1e-4, 2, 1 - 1e-2, 5, 1, 1 - 0.5 * 1e-3]),
    )

    expected_result = np.array([False for _ in range(6)])
    expected_result[[0, 2, 4, 5]] = True

    assert all(
        ModesSelectors.stable_modes(max_distance_from_unity_outside=1e-3)(
            fake_dmd
        )
        == expected_result
    )


def test_stable_modes_inside_only():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array([1 + 1e-4, 2, 1 - 1e-2, 5, 1, 1 - 0.5 * 1e-3]),
    )

    expected_result = np.array([False for _ in range(6)])
    expected_result[[0, 1, 3, 4, 5]] = True

    assert all(
        ModesSelectors.stable_modes(max_distance_from_unity_inside=1e-3)(
            fake_dmd
        )
        == expected_result
    )


def test_stable_modes_errors():
    with raises(ValueError):
        ModesSelectors.stable_modes()
    with raises(ValueError):
        ModesSelectors.stable_modes(
            max_distance_from_unity=1.0e-2,
            max_distance_from_unity_inside=1.0e-3,
        )
    with raises(ValueError):
        ModesSelectors.stable_modes(
            max_distance_from_unity=1.0e-2,
            max_distance_from_unity_outside=1.0e-3,
        )


def test_threshold():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    expected_result = np.array([False for _ in range(6)])
    expected_result[[1, 5]] = True

    assert all(
        ModesSelectors.threshold(1 + 1.0e-3, 2 + 1.0e-10)(fake_dmd)
        == expected_result
    )


def test_compute_integral_contribution():
    np.testing.assert_almost_equal(
        ModesSelectors._compute_integral_contribution(
            np.array([5, 0, 0, 1]), np.array([1, -2, 3, -5, 6])
        ),
        442,
        decimal=1,
    )


def test_integral_contribution():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "dynamics",
        np.array([[i for _ in range(10)] for i in range(4)]),
    )
    setattr(fake_dmd, "modes", np.ones((20, 4)))
    setattr(fake_dmd, "dmd_time", None)
    setattr(fake_dmd, "original_time", None)

    expected_result = np.array([False for _ in range(4)])
    expected_result[[2, 3]] = True

    assert all(
        ModesSelectors.integral_contribution(2)(fake_dmd) == expected_result
    )


def test_integral_contribution_reconstruction():
    dmd = DMD(svd_rank=10)
    dmd.fit(sample_data)
    exp = dmd.reconstructed_data
    select_modes(dmd, ModesSelectors.integral_contribution(2))
    np.testing.assert_array_almost_equal(exp, dmd.reconstructed_data)


def test_stabilize_modes():
    dmd = DMD()
    fake_dmd_operator = FakeDMDOperator()

    eigs = np.array(
        [
            complex(0.3, 0.2),
            complex(0.8, 0.5),
            1,
            complex(1, 1.0e-2),
            2,
            complex(2, 1.0e-2),
        ]
    )
    amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(dmd, "_Atilde", fake_dmd_operator)
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))

    setattr(dmd, "_b", amplitudes)

    stabilize_modes(dmd, 0.8, 1.2)
    np.testing.assert_array_almost_equal(
        dmd.eigs,
        np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5) / abs(complex(0.8, 0.5)),
                1,
                complex(1, 1.0e-2) / abs(complex(1, 1.0e-2)),
                2,
                complex(2, 1.0e-2),
            ]
        ),
    )

    np.testing.assert_array_almost_equal(
        dmd.amplitudes,
        np.array(
            [
                1,
                2 * abs(complex(0.8, 0.5)),
                3,
                4 * abs(complex(1, 1.0e-2)),
                5,
                6,
            ]
        ),
    )


def test_stabilize_modes_index():
    dmd = DMD()
    fake_dmd_operator = FakeDMDOperator()

    eigs = np.array(
        [
            complex(0.3, 0.2),
            complex(0.8, 0.5),
            1,
            complex(1, 1.0e-2),
            2,
            complex(2, 1.0e-2),
        ]
    )
    amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
    setattr(dmd, "_Atilde", fake_dmd_operator)

    setattr(dmd, "_b", amplitudes)

    _, indexes = stabilize_modes(dmd, 0.8, 1.2, return_indexes=True)

    np.testing.assert_array_almost_equal(
        dmd.eigs,
        np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5) / abs(complex(0.8, 0.5)),
                1,
                complex(1, 1.0e-2) / abs(complex(1, 1.0e-2)),
                2,
                complex(2, 1.0e-2),
            ]
        ),
    )

    np.testing.assert_array_almost_equal(
        dmd.amplitudes,
        np.array(
            [
                1,
                2 * abs(complex(0.8, 0.5)),
                3,
                4 * abs(complex(1, 1.0e-2)),
                5,
                6,
            ]
        ),
    )

    np.testing.assert_almost_equal(indexes, [1, 2, 3])


def test_stabilize_modes_index_deepcopy():
    dmd = DMD()
    fake_dmd_operator = FakeDMDOperator()

    eigs = np.array(
        [
            complex(0.3, 0.2),
            complex(0.8, 0.5),
            1,
            complex(1, 1.0e-2),
            2,
            complex(2, 1.0e-2),
        ]
    )
    amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
    setattr(dmd, "_Atilde", fake_dmd_operator)

    setattr(dmd, "_b", amplitudes)

    dmd2, indexes = stabilize_modes(
        dmd, 0.8, 1.2, in_place=False, return_indexes=True
    )

    np.testing.assert_array_almost_equal(
        dmd2.eigs,
        np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5) / abs(complex(0.8, 0.5)),
                1,
                complex(1, 1.0e-2) / abs(complex(1, 1.0e-2)),
                2,
                complex(2, 1.0e-2),
            ]
        ),
    )

    np.testing.assert_array_almost_equal(
        dmd2.amplitudes,
        np.array(
            [
                1,
                2 * abs(complex(0.8, 0.5)),
                3,
                4 * abs(complex(1, 1.0e-2)),
                5,
                6,
            ]
        ),
    )

    np.testing.assert_array_almost_equal(
        dmd.eigs,
        np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5),
                1,
                complex(1, 1.0e-2),
                2,
                complex(2, 1.0e-2),
            ]
        ),
    )

    np.testing.assert_array_almost_equal(
        dmd.amplitudes, np.array([1, 2, 3, 4, 5, 6])
    )

    np.testing.assert_almost_equal(indexes, [1, 2, 3])


# test that the dmd given to ModesTuner is copied with deepcopy
def test_modes_tuner_copy():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    ModesTuner(fake_dmd)._dmds[0].eigs[1] = 0
    assert fake_dmd.eigs[1] == 2


# assert that passing a scalar DMD (i.e. no list) causes ModesTuner to return
# only scalar DMD instances
def test_modes_tuner_scalar_input():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    mt = ModesTuner(fake_dmd, in_place=True)
    assert mt.get() == fake_dmd
    assert isinstance(mt.copy(), FakeDMD)


def test_modes_tuner_list_input():
    class FakeDMD:
        pass

    def cook_fake_dmd():
        fake_dmd = FakeDMD()
        setattr(
            fake_dmd,
            "eigs",
            np.array(
                [
                    complex(1, 1e-4),
                    2,
                    complex(1, 1e-2),
                    5,
                    1,
                    complex(1, 5 * 1e-2),
                ]
            ),
        )
        return fake_dmd

    dmd1 = cook_fake_dmd()
    dmd2 = cook_fake_dmd()

    mt = ModesTuner([dmd1, dmd2], in_place=True)
    assert isinstance(mt.get(), list)
    assert mt.get()[0] == dmd1
    assert mt.get()[1] == dmd2

    assert isinstance(mt.copy(), list)
    assert len(mt.copy()) == 2


def test_modes_tuner_get():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    mtuner = ModesTuner(fake_dmd)

    eigs = mtuner.get().eigs
    mtuner._dmds[0].eigs[1] = 0
    assert eigs[1] == 0


def test_modes_tuner_secure_copy():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    mtuner = ModesTuner(fake_dmd)

    eigs = mtuner.copy().eigs
    mtuner._dmds[0].eigs[1] = 0
    assert eigs[1] == 2


def test_modes_tuner_inplace():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    mtuner = ModesTuner(fake_dmd, in_place=True)
    assert mtuner.get() == fake_dmd

    mtuner._dmds[0].eigs[1] = 0
    assert fake_dmd.eigs[1] == 0


def test_modes_tuner_inplace_list():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )
    fake_dmd2 = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 3, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    mtuner = ModesTuner([fake_dmd, fake_dmd2], in_place=True)
    assert mtuner.get()[0] == fake_dmd
    assert mtuner.get()[1] == fake_dmd2

    mtuner._dmds[0].eigs[1] = 0
    assert fake_dmd.eigs[1] == 0


def test_modes_tuner_select_raises():
    class FakeDMD:
        pass

    fake_dmd = FakeDMD()
    setattr(
        fake_dmd,
        "eigs",
        np.array(
            [complex(1, 1e-4), 2, complex(1, 1e-2), 5, 1, complex(1, 5 * 1e-2)]
        ),
    )

    with raises(ValueError):
        ModesTuner(fake_dmd).select("ciauu")
    with raises(ValueError):
        ModesTuner(fake_dmd).select(2)


def test_modes_tuner_select():
    fake_dmd_operator = FakeDMDOperator()
    fake_dmd = DMD()

    eigs = np.array([1 + 1e-4, 2, 1 - 1e-2, 5, 1, 1 - 0.5 * 1e-3])

    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "_Lambda", np.zeros(len(eigs)))
    # these are DMD eigenvectors, but we do not care in this test
    setattr(fake_dmd_operator, "_eigenvectors", np.zeros((1, len(eigs))))
    setattr(fake_dmd_operator, "_modes", np.zeros((1, len(eigs))))
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
    setattr(fake_dmd, "_b", np.zeros(len(eigs)))

    setattr(fake_dmd, "_Atilde", fake_dmd_operator)

    mtuner = ModesTuner(fake_dmd)
    mtuner.select("stable_modes", max_distance_from_unity=1e-3)
    dmd = mtuner.get()

    assert len(dmd.eigs) == 3
    assert len(dmd.amplitudes) == 3
    assert dmd.modes.shape[1] == 3


def test_modes_tuner_stabilize():
    dmd = DMD()
    fake_dmd_operator = FakeDMDOperator()

    eigs = np.array(
        [
            complex(0.3, 0.2),
            complex(0.8, 0.5),
            1,
            complex(1, 1.0e-2),
            2,
            complex(2, 1.0e-2),
        ]
    )
    amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

    setattr(fake_dmd_operator, "_eigenvalues", eigs)
    setattr(fake_dmd_operator, "eigenvalues", eigs)
    setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
    setattr(dmd, "_Atilde", fake_dmd_operator)

    setattr(dmd, "_b", amplitudes)

    mtuner = ModesTuner(dmd)
    mtuner.stabilize(inner_radius=0.8, outer_radius=1.2)
    dmd = mtuner.get()

    np.testing.assert_array_almost_equal(
        dmd.eigs,
        np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5) / abs(complex(0.8, 0.5)),
                1,
                complex(1, 1.0e-2) / abs(complex(1, 1.0e-2)),
                2,
                complex(2, 1.0e-2),
            ]
        ),
    )

    np.testing.assert_array_almost_equal(
        dmd.amplitudes,
        np.array(
            [
                1,
                2 * abs(complex(0.8, 0.5)),
                3,
                4 * abs(complex(1, 1.0e-2)),
                5,
                6,
            ]
        ),
    )


def test_modes_tuner_stabilize_multiple():
    def cook_fake_dmd():
        dmd = DMD()
        fake_dmd_operator = FakeDMDOperator()

        eigs = np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5),
                1,
                complex(1, 1.0e-2),
                2,
                complex(2, 1.0e-2),
            ]
        )
        amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

        setattr(fake_dmd_operator, "_eigenvalues", eigs)
        setattr(fake_dmd_operator, "eigenvalues", eigs)
        setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
        setattr(dmd, "_Atilde", fake_dmd_operator)

        setattr(dmd, "_b", amplitudes)

        return dmd

    dmd1 = cook_fake_dmd()
    dmd2 = cook_fake_dmd()
    dmd3 = cook_fake_dmd()

    mtuner = ModesTuner([dmd1, dmd2, dmd3])
    mtuner.stabilize(inner_radius=0.8, outer_radius=1.2)
    dmds = mtuner.get()

    assert isinstance(dmds, list)

    for dmd in dmds:
        np.testing.assert_array_almost_equal(
            dmd.eigs,
            np.array(
                [
                    complex(0.3, 0.2),
                    complex(0.8, 0.5) / abs(complex(0.8, 0.5)),
                    1,
                    complex(1, 1.0e-2) / abs(complex(1, 1.0e-2)),
                    2,
                    complex(2, 1.0e-2),
                ]
            ),
        )

        np.testing.assert_array_almost_equal(
            dmd.amplitudes,
            np.array(
                [
                    1,
                    2 * abs(complex(0.8, 0.5)),
                    3,
                    4 * abs(complex(1, 1.0e-2)),
                    5,
                    6,
                ]
            ),
        )


def test_modes_tuner_subset():
    def cook_fake_dmd():
        dmd = DMD()
        fake_dmd_operator = FakeDMDOperator()

        eigs = np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5),
                1,
                complex(1, 1.0e-2),
                2,
                complex(2, 1.0e-2),
            ]
        )
        amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

        setattr(fake_dmd_operator, "_eigenvalues", eigs)
        setattr(fake_dmd_operator, "eigenvalues", eigs)
        setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
        setattr(dmd, "_Atilde", fake_dmd_operator)

        setattr(dmd, "_b", amplitudes)

        return dmd

    dmd1 = cook_fake_dmd()
    dmd2 = cook_fake_dmd()
    dmd3 = cook_fake_dmd()

    mtuner = ModesTuner([dmd1, dmd2, dmd3], in_place=True)
    assert len(mtuner.subset([0, 2]).get()) == 2
    assert mtuner.subset([0, 2]).get()[0] == dmd1
    assert mtuner.subset([0, 2]).get()[1] == dmd3

    mtuner = ModesTuner([dmd1, dmd2, dmd3], in_place=False)
    assert len(mtuner.subset([0, 2]).get()) == 2
    assert mtuner.subset([0, 2]).get()[0] == mtuner._dmds[0]
    assert mtuner.subset([0, 2]).get()[1] == mtuner._dmds[2]


def test_modes_tuner_stabilize_multiple_subset():
    def cook_fake_dmd():
        dmd = DMD()
        fake_dmd_operator = FakeDMDOperator()

        eigs = np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5),
                1,
                complex(1, 1.0e-2),
                2,
                complex(2, 1.0e-2),
            ]
        )
        amplitudes = np.array([1, 2, 3, 4, 5, 6], dtype=complex)

        setattr(fake_dmd_operator, "_eigenvalues", eigs)
        setattr(fake_dmd_operator, "eigenvalues", eigs)
        setattr(fake_dmd_operator, "modes", np.zeros((1, len(eigs))))
        setattr(dmd, "_Atilde", fake_dmd_operator)

        setattr(dmd, "_b", amplitudes)

        return dmd

    dmd1 = cook_fake_dmd()
    dmd2 = cook_fake_dmd()
    dmd3 = cook_fake_dmd()

    mtuner = ModesTuner([dmd1, dmd2, dmd3])
    mtuner.subset([0, 2]).stabilize(inner_radius=0.8, outer_radius=1.2)
    dmds = mtuner.get()

    assert len(dmds) == 3

    for i in range(3):
        if i == 1:
            continue
        np.testing.assert_array_almost_equal(
            dmds[i].eigs,
            np.array(
                [
                    complex(0.3, 0.2),
                    complex(0.8, 0.5) / abs(complex(0.8, 0.5)),
                    1,
                    complex(1, 1.0e-2) / abs(complex(1, 1.0e-2)),
                    2,
                    complex(2, 1.0e-2),
                ]
            ),
        )
        np.testing.assert_array_almost_equal(
            dmds[i].amplitudes,
            np.array(
                [
                    1,
                    2 * abs(complex(0.8, 0.5)),
                    3,
                    4 * abs(complex(1, 1.0e-2)),
                    5,
                    6,
                ]
            ),
        )

    np.testing.assert_array_almost_equal(
        dmds[1].eigs,
        np.array(
            [
                complex(0.3, 0.2),
                complex(0.8, 0.5),
                1,
                complex(1, 1.0e-2),
                2,
                complex(2, 1.0e-2),
            ]
        ),
    )
    np.testing.assert_array_almost_equal(
        dmds[1].amplitudes, np.array([1, 2, 3, 4, 5, 6], dtype=complex)
    )


def test_modes_tuner_index_scalar_dmd_raises():
    def stable_modes(dmd_object):
        toll = 1e-3
        return np.abs(np.abs(dmd_object.eigs) - 1) < toll

    dmd = DMD(svd_rank=10)
    dmd.fit(sample_data)

    with raises(ValueError):
        ModesTuner(dmd).subset([0])


def test_modes_tuner_selectors():
    assert selectors["module_threshold"] == ModesSelectors.threshold
    assert selectors["stable_modes"] == ModesSelectors.stable_modes
    assert (
        selectors["integral_contribution"]
        == ModesSelectors.integral_contribution
    )


@pytest.mark.parametrize(
    "dmd",
    [
        param(CDMD(svd_rank=-1), id="CDMD"),
        param(DMD(svd_rank=-1), id="DMD"),
        param(DMDc(svd_rank=-1), id="DMDc"),
        param(FbDMD(svd_rank=-1), id="FbDMD"),
        param(HankelDMD(svd_rank=-1, d=3), id="HankelDMD"),
        param(HODMD(svd_rank=-1, d=3), id="HODMD"),
    ],
)
def test_modes_selector_all_dmd_types(dmd):
    print(
        "--------------------------- {} ---------------------------".format(
            type(dmd)
        )
    )
    if isinstance(dmd, ParametricDMD):
        repeated = np.repeat(sample_data[None], 10, axis=0)
        dmd.fit(repeated + np.random.rand(*repeated.shape), np.ones(10))
    elif isinstance(dmd, DMDc):
        snapshots = np.array(
            [[4, 2, 1, 0.5, 0.25], [7, 0.7, 0.07, 0.007, 0.0007]]
        )
        u = np.array([-4, -2, -1, -0.5])
        B = np.array([[1, 0]]).T
        dmd.fit(snapshots, u, B)
    else:
        dmd.fit(sample_data)

    ModesTuner(dmd, in_place=True).select(
        "integral_contribution", n=3
    ).stabilize(1 - 1.0e-3)
    assert True
